# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import typing
import stat
import tempfile
import logging
import contextlib
import shutil
import fs
from stat import filemode
from datetime import datetime
from os import F_OK, R_OK, W_OK
from typing import Optional, Union, Text, Any, List
from fs import open_fs, mirror, errors, subfs, base
from fs.time import datetime_to_epoch
from fs.mode import Mode
from fs.wrapfs import WrapFS
from fs.permissions import Permissions
from fs.osfs import Info
from conpot.helpers import months_map
from types import FunctionType
from conpot.core.fs_utils import _custom_conpot_file, SubAbstractFS, copy_files, FilesystemError
from conpot.core.fs_utils import FSOperationNotPermitted

_F = typing.TypeVar('_F', bound='FS', covariant=True)

logger = logging.getLogger(__name__)


# ---------------------------------------------------
# Regarding Permissions:
# ---------------------------------------------------
# For directories:
#   - Read bit = You can read the names on the list.
#   - Write bit = You can {add,rename,delete} names on the list IF the execute bit is set too.
#   - Execute bit = You can make this directory your working directory.
# ---------------------------------------------------
# For files:
#   - Read bit = Grants the capability to read, i.e., view the contents of the file.
#   - Write bit = Grants the capability to modify, or remove the content of the file.
#   - Execute bit = User with execute permissions can run a file as a program.
# ---------------------------------------------------

class AbstractFS(WrapFS):
    """
    AbstractFS distinguishes between "real" filesystem paths and "virtual" ftp paths emulating a UNIX chroot jail
    where the user can not escape its home directory (example: real "/home/user" path will be seen as "/" by the client)

    This class exposes common fs wrappers around all os.* calls involving operations against the filesystem like
    creating files or removing directories (such as listdir etc.)

    *Implementation Note:* When doing I/O - Always with the check_access and set_access context managers for safe
    operations.
    """

    def __init__(self,
                 src_path: str,
                 create_mode: int = 0o777,      # Default file system permissions.
                 temp_dir: Union[str, None]=None,
                 identifier: Optional[str]='__conpot__',
                 auto_clean: Optional[bool]=True,
                 ignore_clean_errors: Optional[bool]=True) -> None:
        self._cwd = self.getcwd()  # keep track of the current working directory
        self._cache = {}   # Storing all cache of the file system
        self.identifier = identifier.replace('/', '-')
        self._auto_clean = auto_clean
        self._ignore_clean_errors = ignore_clean_errors
        self.temp_dir = temp_dir
        self._cleaned = False
        self.built_cache = False
        # Create our file system
        self._temp_dir = tempfile.mkdtemp(
            prefix=(self.identifier or "ConpotTempFS"),
            dir=self.temp_dir
        )
        # open various filesystems that would be used by Conpot
        try:
            self.vfs = open_fs(self._temp_dir)
            super(AbstractFS, self).__init__(self.vfs)
        except fs.errors.FSError as fs_err:
            logger.exception('File System exception occurred! {}'.format(fs_err))
        # Copy all files from src_path into our file system
        logger.info('Initializing Virtual File System at {}. Source specified : {}\n Please wait while the '
                    'system copies all specified files'.format(self._temp_dir, src_path))
        self.utime = self.settimes  # utime maps to settimes
        # keep records related to users and groups
        self.default_uid = 0
        self.default_gid = 0
        self.default_perms = create_mode
        self._users = {
            0: {'user': 'root'}
        }
        self._grps = {
            0: {'group': 'root'}
        }
        # simple dictionary linking users to groups ->
        self._user_grps = {
            0: {0}      # --> gid: set(uids)
        }
        self._initialize_fs(src_path=src_path)
        # fixme: kind of hack-ish. Find the correct way of doing this.
        self._wrap_fs._meta['supports_rename'] = False

    def norm_path(self, path):
        path = '/' if path == '.' else path
        try:
            _path = self.validatepath(self._cwd + path) if self._cwd not in path else self.validatepath(path)
            return _path
        except fs.errors.FSError:
            logger.debug('Could not validate path: {}'.format(path))
            raise FilesystemError('Could not validate path: {}'.format(path))

    def _initialize_fs(self, src_path: str) -> None:
            """
            Copies all data into Conpot's created fs folder and builds up the cache.
            :param src_path: FS URLS
            """
            # copy all contents from the source path the filesystem.
            src_fs = open_fs(src_path)
            logger.debug('Building up file system: copying contents from the source path {}'.format(src_path))
            with src_fs.lock():
                mirror.mirror(src_fs=src_fs, dst_fs=self.vfs)
                self._cache.update({path: info for path, info in self.walk.info(namespaces=['basic', 'access',
                                                                                            'details', 'stat'])})
                self._cache['/'] = self._wrap_fs.getinfo('/', namespaces=['basic', 'access', 'details', 'stat', 'link'])
                self.chown('/', self.default_uid, self.default_gid, recursive=True)
                self.chmod('/', self.default_perms, recursive=True)
                self.built_cache = True   # FS has been built. Now all info must be accessed from cache.
                src_fs.close()
            del src_fs

    def __str__(self):
        return "<Conpot AbstractFS '{}'>".format(self._temp_dir)

    __repr__ = __str__

    @property
    def users(self):
        return self._users

    @property
    def groups(self):
        return self._grps

    @property
    def user_groups(self):
        """gid: {set of uid of users.}"""
        return self._user_grps

    def getmeta(self, namespace="standard"):
        self.check()
        meta = self.delegate_fs().getmeta(namespace=namespace)
        meta['supports_rename'] = False
        return meta

    # ------- context managers for easier handling of fs

    @contextlib.contextmanager
    def check_access(self, path=None, user=None, perms=None):
        """
            Checks whether the current user has permissions to do a specific operation. Raises FSOperationNotPermitted
            exception in case permissions are not satisfied.
            Handy utility to check whether the user with uid provided has permissions specified. Examples:

            >>> import conpot.core as conpot_core
            >>> _vfs, _ = conpot_core.get_vfs('ftp')
            >>> with _vfs.check_access(path='/', user=13, perms='rwx'):
            >>>     _vfs.listdir('/')

            >>> with _vfs.check_access(path='/', user=45, perms='w'):
            >>>     with _vfs.open('/test', mode='wb') as _file:
            >>>         _file.write(b'Hello World!')
        """
        if not self.access(path=path, name_or_id=user, required_perms=perms):
            raise FSOperationNotPermitted('User {} does not have required permission to file/path: {}'.format(
                user, path)
            )
        else:
            logger.debug('File/Dir {} has the requested params : {}'.format(path, (user, perms)))
            self.setinfo(path, {})
            yield
            if self.vfs.isfile(path):
                logger.debug('yield file: {} after requested access.'.format(path))
            elif self.vfs.isdir(path):
                logger.debug('yield dir: {} after requested access.'.format(path))
            else:
                logger.debug('yield unknown type: {} after requested access.'.format(path))

    # -----------------------------------------------------------
    # Custom "setter" methods overwriting behaviour FS library methods
    # We need to update our cache in such cases.
    # -----------------------------------------------------------
    def setinfo(self, path, info):
        """
        Higher level function to directly change values in the file system. Dictionary specified here changes cache
        values.
        :param path: path of the file that is to be changed
        :param info: Raw Info object. Please check pyfilesystem2's docs for more info.
        """
        assert path and isinstance(path, str)
        path = self.norm_path(path)
        if 'lstat' not in info:
            try:
                if 'details' in info:
                    details = info['details']
                    if 'accessed' in details or 'modified' in details:
                        return self._wrap_fs.setinfo(path, info)
            finally:
                try:
                    assert self._cache[path]
                except (AssertionError, KeyError):
                    # This is the first time we have seen this file. Let us create this entry.
                    logger.debug('Creating cache for file/directory : {}'.format(path))
                    self._cache[path] = self._wrap_fs.getinfo(path, namespaces=['basic', 'access', 'details', 'stat',
                                                                                'link'])
                # update the 'accessed' and 'modified' time.
                self.settimes(path)
                if 'access' in info:
                    access = info['access']
                    if 'permissions' in access:
                        self._cache[path].raw['access']['permissions'] = access['permissions']
                        self._cache[path].raw['details']['metadata_changed'] = fs.time.datetime_to_epoch(datetime.now())
                    if 'user' in access or 'uid' in access:
                        try:
                            if 'user' in access or ('user' in access and 'uid' in access):
                                self._cache[path].raw['access']['user'] = access['user']
                                [_uid] = [key for key, value in self._users.items()
                                          if value == {'user': access['user']}]
                                self._cache[path].raw['access']['uid'] = _uid
                                self._cache[path].raw['details']['metadata_changed'] = fs.time.datetime_to_epoch(
                                    datetime.now()
                                )
                            else:
                                # Must be 'uid' that is available.
                                _uid = int(access['uid'])   # type: ignore
                                self._cache[path].raw['access']['uid'] = _uid
                                self._cache[path].raw['access']['user'] = self._users[_uid]['user']
                                self._cache[path].raw['details']['metadata_changed'] = fs.time.datetime_to_epoch(
                                    datetime.now()
                                )
                        except (TypeError, AssertionError, KeyError):
                            raise
                    if 'group' in access or 'gid' in access:
                        try:
                            if 'group' in access or ('group' in access and 'gid' in access):
                                self._cache[path].raw['access']['group'] = access['group']
                                [_gid] = [key for key, value in self._grps.items()
                                          if value == {'group': access['group']}]
                                self._cache[path].raw['access']['gid'] = _gid
                                self._cache[path].raw['details']['metadata_changed'] = fs.time.datetime_to_epoch(
                                    datetime.now()
                                )
                            else:
                                # Must be 'gid' that is available.
                                _gid = int(access['gid'])    # type: ignore
                                self._cache[path].raw['access']['gid'] = _gid
                                self._cache[path].raw['access']['group'] = self._grps[_gid]['group']
                                self._cache[path].raw['details']['metadata_changed'] = fs.time.datetime_to_epoch(
                                    datetime.now()
                                )
                        except (TypeError, AssertionError, KeyError):
                            raise
        else:
            raise FilesystemError('lstat is not currently supported!')

    def makedir(self,
                path,               # type: Text
                permissions=None,   # type: Optional[int]
                recreate=True       # type: bool
                ):
        # make a directory in the file system. Also, update the cache about the directory.
        _path = self.norm_path(path)
        # we always want to overwrite a directory if it already exists.
        recreate = True if recreate is False else recreate
        perms = permissions
        fs_err = None
        try:
            super(AbstractFS, self).makedir(_path, permissions=None, recreate=recreate)
        except fs.errors.FSError as err:
            fs_err = err
        finally:
            if not fs_err:
                dir_perms = perms if perms else self.default_perms
                dir_cache = {
                    'access': {
                        'permissions': Permissions.create(dir_perms),
                        'uid': self.default_uid,
                        'gid': self.default_gid
                    }
                }
                logger.debug('Created directory {}'.format(_path))
                self.setinfo(_path, info=dir_cache)
            else:
                raise fs_err

    def removedir(self, path, rf=True):
        """Remove a directory from the file system.
        :param path: directory path
        :param rf: remove directory recursively and forcefully. This removes directory even if there is any data
        in it. If set to False, an exception would be raised
        """
        # removing a directory and finally block would clear the local cache.
        _path = self.norm_path(path)
        fs_err = None
        try:
            super(AbstractFS, self).removedir(_path)
        except fs.errors.FSError as err:
            fs_err = err
        finally:
            if not fs_err:
                rm_dir = self._cache.pop(_path)
                logger.debug('Removed directory {}'.format(rm_dir))
            else:
                if isinstance(fs_err, fs.errors.DirectoryNotEmpty) and rf is True:
                    # delete the contents for the directory recursively
                    self._wrap_fs.removetree(_path)
                    # delete the all the _cache files in the directory.
                    _files = [i for i in self._cache.keys() if _path in i]
                    for _f in _files:
                        file = self._cache.pop(_f)
                        logger.debug('Removing file : {}'.format(repr(file)))
                else:
                    raise fs_err

    def remove(self, path):
        """Remove a file from the file system."""
        _path = self.norm_path(path)
        fs_err = None
        try:
            super(AbstractFS, self).remove(_path)
        except fs.errors.FSError as err:
            fs_err = err
        finally:
            if not fs_err:
                rm_file = self._cache.pop(_path)
                logger.debug('Removed file {}'.format(rm_file))
            else:
                raise fs_err

    def openbin(self, path, mode='r', buffering=-1, **options):
        """
        Open a file in the ConpotFS in binary mode.
        """
        logger.debug('Opening file {} with mode {}'.format(path, mode))
        _path = self.norm_path(path)
        _bin_mode = Mode(mode).to_platform_bin()
        _bin_mode = _bin_mode.replace("t", "") if "t" in _bin_mode else _bin_mode
        _parent_fs = self.delegate_fs()
        self.check()
        binary_file = _custom_conpot_file(file_system=self,
                                          parent_fs=_parent_fs,
                                          path=_path,
                                          mode=_bin_mode,
                                          encoding=None)
        return binary_file

    def open(self,
             path,                      # type: Text
             mode='r',                  # type: Text
             buffering=-1,              # type: int
             encoding=None,             # type: Optional[Text]
             newline='',                # type: Text
             line_buffering=False,      # type: bool
             **options                  # type: Any
             ):
        _open_mode = Mode(mode)
        base.validate_open_mode(mode)
        self.check()
        _path = self.norm_path(path)
        _parent_fs = self.delegate_fs()
        _encoding = encoding or "utf-8"
        file = _custom_conpot_file(file_system=self,
                                   parent_fs=_parent_fs,
                                   path=_path,
                                   mode=_open_mode.to_platform(),
                                   buffering=buffering,
                                   encoding=encoding,
                                   newline=newline,
                                   line_buffering=line_buffering)
        return file

    def setbinfile(self, path, file):
        with self.openbin(path, "wb") as dst_file:
            copy_files(file, dst_file)
        self.setinfo(path, {})

    def move(self, src_path, dst_path, overwrite=False):
        if self.getinfo(src_path).is_dir:
            raise fs.errors.FileExpected(src_path)
        with self.openbin(src_path, "rb") as read_file:
            with self.openbin(dst_path, "wb") as dst_file:
                copy_files(read_file, dst_file)
        self.setinfo(src_path, {})
        self.setinfo(dst_path, {})
        self.remove(src_path)

    def copy(self, src_path, dst_path, overwrite=False):
        if self.getinfo(src_path).is_dir:
            raise fs.errors.FileExpected(src_path)
        with self.openbin(src_path, "rb") as read_file:
            with self.openbin(dst_path, "wb") as dst_file:
                copy_files(read_file, dst_file)
        self.setinfo(src_path, {})
        self.setinfo(dst_path, {})

    # -----------------------------------------------------------
    # Custom "getter" methods overwriting behaviour FS library methods
    # Data is retrieved from the cached file-system.
    # -----------------------------------------------------------

    def opendir(self, path, factory=SubAbstractFS):
        return super(AbstractFS, self).opendir(path, factory=factory)

    def settimes(self, path, accessed=None, modified=None):
        if accessed or modified:
            self.delegate_fs().settimes(path, accessed, modified)
        self._cache[path].raw['details']['accessed'] = \
            fs.time.datetime_to_epoch(super(AbstractFS, self).getinfo(path, namespaces=['details']).accessed)
        self._cache[path].raw['details']['modified'] = \
            fs.time.datetime_to_epoch(super(AbstractFS, self).getinfo(path, namespaces=['details']).modified)

    def getinfo(self, path: str, get_actual: bool = False, namespaces=None):
        if get_actual or (not self.built_cache):
            return self._wrap_fs.getinfo(path, namespaces)
        else:
            try:
                # ensure that the path starts with '/'
                if path[0] != '/':
                    path = '/' + path
                info = {'basic': self._cache[path].raw['basic']}
                if namespaces is not None:
                    if 'details' in namespaces:
                        info['details'] = self._cache[path].raw['details']
                    if 'stat' in namespaces:
                        stat_cache = {
                            'st_uid': self._cache[path].raw['access']['uid'],
                            'st_gid': self._cache[path].raw['access']['gid'],
                            'st_atime': self._cache[path].raw['details']['accessed'],
                            'st_mtime': self._cache[path].raw['details']['modified'],
                            # TODO: Fix these to appropriate values
                            'st_mtime_ns': None,
                            'st_ctime_ns': None,
                            'st_ctime': None,
                        }
                        if isinstance(self._cache[path].raw['access']['permissions'], list):
                            stat_cache['st_mode'] = Permissions(self._cache[path].raw['access']['permissions']).mode
                        else:
                            stat_cache['st_mode'] = self._cache[path].raw['access']['permissions'].mode
                        self._cache[path].raw['stat'].update(stat_cache)
                        info['stat'] = self._cache[path].raw['stat']
                        # Note that we won't be keeping tabs on 'lstat'
                    if 'lstat' in namespaces:
                        info['lstat'] = self._cache[path].raw['lstat']
                        info['lstat'] = self._cache[path].raw['lstat']
                    if 'link' in namespaces:
                        info['link'] = self._cache[path].raw['link']
                    if 'access' in namespaces:
                        info['access'] = self._cache[path].raw['access']
                return Info(info)
            except KeyError:
                raise FilesystemError

    def listdir(self, path):
        logger.debug('Listing contents from directory: {}'.format(self.norm_path(path)))
        self.setinfo(self.norm_path(path), {})
        return super(AbstractFS, self).listdir(self.norm_path(path))

    def getfile(self, path, file, chunk_size=None, **options):
        # check where there exists a copy in the cache
        if self.exists(self.norm_path(path)) and self.norm_path(path) in self._cache.keys():
            self.setinfo(self.norm_path(path), {})
            return self._wrap_fs.getfile(self.norm_path(path), file, chunk_size, **options)
        else:
            raise FilesystemError('Can\'t get. File does not exist!')

    def __del__(self):
        self._wrap_fs.close()
        if self._auto_clean:
            self.clean()

    # -----------------------------------------------------------
    # Methods defined for our needs.
    # -----------------------------------------------------------

    def create_jail(self, path):
        """Returns chroot jail sub system for a path"""
        logger.debug('Creating jail for path: {}'.format(path))
        return self.opendir(path)

    def getcwd(self):
        return '/'

    def take_snapshot(self):
        """Take snapshot of entire filesystem.
        :rtype: dict
        """
        return {'date-time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"), 'snapshot-data': self._cache}

    def register_user(self, name: str, uid: int) -> None:
        """Store all user related data for the file system."""
        assert name and isinstance(name, str)
        self._users[uid] = {'user': name}
        # let us check for duplicate usernames/group names
        if len(set([v['user'] for k, v in self._users.items()])) != len(self._users.keys()):
            _uname = self._users.pop(uid)['user']
            raise FilesystemError('Can\'t add users with duplicate uname: {}.'.format(_uname))

    def create_group(self,  name: str, gid: int) -> None:
        """
        Store all group related data for the file system.
        :param name: Name of the group
        :param gid: gid of the group
        """
        assert name and isinstance(name, str)
        self._grps[gid] = {'group': name}
        if len(set([v['group'] for k, v in self._grps.items()])) != len(self._grps.keys()):
            _gname = self._grps.pop(gid)
            raise FilesystemError('Can\'t create groups with duplicate names: {}.'.format(_gname))

    def add_users_to_group(self, gid: int, uids: List) -> None:
        """Add list of users to an existing group
        :param gid: Group id of the group.
        :param uids: List of registers users that belong to this group
        """
        try:
            assert gid in self._grps.keys()
            for i in uids:
                if i not in self._users.keys():
                    raise AssertionError
            _uids = set(uids)
            if gid in self._user_grps.keys():
                self._user_grps[gid] += _uids
            else:
                self._user_grps[gid] = _uids
        except AssertionError:
            raise FilesystemError('uid/gid does not exist in the file system. Please register it via create_group/'
                                  'register_user method.')

    def chown(self, fs_path: str, uid: int, gid: int, recursive: Optional[bool]=False) -> None:
        """Change the owner of a specified file. Wrapper for os.chown
        :param fs_path: path or directory in the VFS where chown would be executed.
        :param uid: The `uid` of the user. **User must be a registered user on the filesystem or an exception would be
        thrown.
        :param gid: The `gid` of the group **Group must be a registered group on the filesystem or an exception would be
        thrown.
        :param recursive: If the given path is directory, then setting the recursive option to true would walk down the
        tree and recursive change permissions in the cache.

        ** `fs_path` needs to be the absolute path w.r.t to the vfs. If you are in a sub file system, please use
        `subvfs.getcwd()` to get the current directory. **
        """
        path = self.norm_path(fs_path)
        try:
            assert isinstance(uid, int) and isinstance(gid, int)
        except AssertionError:
            logger.exception('Integers expected got {} - {}'.format(uid, gid))
        if self.isdir(path) or self.isfile(path):
            assert self._grps[gid] and self._users[uid]
            chown_cache = {
                'access': {
                    'user': self._users[uid]['user'],
                    'uid': self._users[uid],
                    'group': self._grps[gid]['group'],
                    'gid': self._grps[gid]
                }
            }
            if self.isdir(path) and recursive:
                if self.norm_path(path) is not '/':
                    self.setinfo(path, chown_cache)
                sub_dir = self.opendir(path)
                for _path, _ in sub_dir.walk.info():
                    assert self._cache[self.norm_path(path + _path)]
                    self.setinfo(path + _path, chown_cache)
                sub_dir.close()
            else:
                self.setinfo(path, chown_cache)

        else:
            # TODO: map this to the actual output of os.chown
            raise FilesystemError('File not found for chown')

    def clean(self):
        """Clean (delete) temporary files created by this filesystem."""
        if self._cleaned:
            return

        try:
            logger.info('Shutting down File System. Cleaning directories at {}'.format(self._temp_dir))
            shutil.rmtree(self._temp_dir)
        except Exception as error:
            if not self._ignore_clean_errors:
                raise errors.OperationFailed(
                    msg="failed to remove temporary directory",
                    exc=error
                )
        self._cleaned = True

    @property
    def root(self):
        """The root directory - where the filesystem is stored"""
        return self._temp_dir

    def stat(self, path):
        """Perform a stat() system call on the given path.
        :param path: (str) must be protocol relative path
        """
        assert path, isinstance(path, str)
        self.setinfo(self.norm_path(path), {})
        return self.getinfo(path, namespaces=['stat']).raw['stat']

    def readlink(self, path):
        """Perform a readlink() system call. Return a string representing the path to which a symbolic link points.
        :param path: (str) must be protocol relative path
        """
        assert path, isinstance(path, str)
        self.setinfo(self.norm_path(path), {})
        return self.getinfo(path, get_actual=True, namespaces=['link']).raw['link']['target']

    def format_list(self, basedir, listing):
        """
        Return an iterator object that yields the entries of given directory emulating the "/bin/ls -lA" UNIX command
        output.
        This is how output should appear:
        -rw-rw-rw-   1 owner   group    7045120 Sep 02  3:47 music.mp3
        drwxrwxrwx   1 owner   group          0 Aug 31 18:50 e-books
        -rw-rw-rw-   1 owner   group        380 Sep 02  3:40 module.py

        :param basedir: (str) must be protocol relative path
        :param listing: (list) list of files to needed for output.
        """
        assert isinstance(basedir, str), basedir
        basedir += '/' if basedir[-1:] != '/' else basedir
        now = time.time()
        for basename in listing:
            file = self.norm_path(basedir + basename)  # for e.g. basedir = '/' and basename = test.png.
            # So file is '/test.png'
            try:
                st = self.stat(file)
            except (fs.errors.FSError, FilesystemError):
                raise
            permission = filemode(Permissions.create(st['st_mode']).mode)
            if self.isdir(file):
                permission = permission.replace('?', 'd')
            elif self.isfile(file):
                permission = permission.replace('?', '-')
            elif self.islink(file):
                permission = permission.replace('?', 'l')
            nlinks = st['st_nlink']
            size = st['st_size']  # file-size
            uname = self.getinfo(path=file, namespaces=['access']).user
            # |-> pwd.getpwuid(st['st_uid']).pw_name would fetch the user_name of the actual owner of these files.
            gname = self.getinfo(path=file, namespaces=['access']).group
            # |-> grp.getgrgid(st['st_gid']).gr_name would fetch the user_name of the actual of these files.
            mtime = time.gmtime(fs.time.datetime_to_epoch(self.getinfo(file, namespaces=['details']).modified))
            if (now - st['st_mtime']) > (180 * 24 * 60 * 60):
                fmtstr = "%d  %Y"
            else:
                fmtstr = "%d %H:%M"
            mtimestr = "%s %s" % (months_map[mtime.tm_mon], time.strftime(fmtstr, mtime))
            if (st['st_mode'] & 61440) == stat.S_IFLNK:
                # if the file is a symlink, resolve it, e.g.  "symlink -> realfile"
                basename = basename + " -> " + self.readlink(file)
                # formatting is matched with proftpd ls output
            line = "%s %3s %-8s %-8s %8s %s %s\r\n" % (permission, nlinks, uname, gname, size, mtimestr, basename)
            yield line

    def getmtime(self, path):
        """Return the last modified time as a number of seconds since the epoch."""
        self.setinfo(self.norm_path(path), {})
        return self.getinfo(path, namespaces=['details']).modified

    # FIXME: refactor to os.access. Mode is missing from the params
    def access(self, path: str, name_or_id: Union[int, str]=None, required_perms: str=None):
        """
            Returns bool w.r.t  the a user/group has permissions to read/write/execute a file.
            This is a wrapper around os.access. But it would accept name or id instead of of just ids.
            Also it can accept required permissions in the form of strings rather than os.F_OK, os.R_OK, os.W_OK etc.

            *Implementation Note*: First we would check whether the current user has the required permissions. If not,
            then we check the group to which this user belongs to. Finally if the user's group also does not meet the
            perms we check for other permissions.
        """
        try:
            _path = self.norm_path(path)
            _perms = self.getinfo(_path, namespaces=['access']).permissions
            _uid = self.getinfo(_path, namespaces=['access']).uid
            _gid = self.getinfo(_path, namespaces=['access']).gid
            if isinstance(required_perms, int):
                if required_perms == F_OK:
                    return True
                elif required_perms == R_OK:
                    required_perms = 'r'
                elif required_perms == W_OK:
                    required_perms = 'w'
            # first we need to find the uid - in case username is provided instead of uid.
            if isinstance(name_or_id, str):
                # must be username or group name
                # fetch the uid/gid of that uname/gname
                [_id] = [k for k, v in self._users.items() if v == {'user': name_or_id}]
            else:
                _id = name_or_id
            # find the gid of this user.
            _grp_id = None
            # FIXME: The above operation can cause incorrect results if one user belongs to more than one group.
            for key, values in self._user_grps.items():
                if _id in values:
                    _grp_id = key
            if _id is not None:
                if _id == _uid:
                    # provided id is the owner
                    return all([_perms.check('u_' + i) for i in list(required_perms)])
                elif _grp_id and (_grp_id == _gid):
                    # provided id is not the owner but belongs to that grp.
                    # That means we would check it's group permissions.
                    return all([_perms.check('g_' + i) for i in list(required_perms)])
                else:
                    # id not equal to either in uid/gid
                    # check other permissions
                    return all([_perms.check('o_' + i) for i in list(required_perms)])
        except (ValueError, AssertionError, KeyError, fs.errors.FSError) as err:
            logger.info('Exception has occurred while doing fs.access: {}'.format(err))
            logger.info('Returning False to avoid conpot crash')
            return False

    def get_permissions(self, path):
        """Get permissions for a particular user on a particular file/directory in 'rwxrx---' format"""
        _path = self.norm_path(path)
        self.setinfo(self.norm_path(path), {})
        _perms = self.getinfo(_path, namespaces=['access']).permissions
        return _perms.as_str()

    def chmod(self, path: str, mode: oct, recursive: bool = False) -> None:
        """Change file/directory mode.
        :param path: Path to be modified.
        :param mode: Operating-system mode bitfield. Must be in octal's form.
        Eg: chmod with (mode=0o755) = Permissions(user='rwx', group='rx', other='rx')
        :param recursive: If the path is directory, setting recursive to true would change permissions to sub folders
        and contained files.
        :type recursive: bool
        """
        assert (isinstance(mode, str) or isinstance(mode, int))
        if isinstance(mode, str):
            # convert mode to octal
            mode = int(mode, 8)
        chmod_cache_info = {
                    'access': {
                        'permissions': Permissions.create(mode)
                    }
                }
        if self.isdir(path) and recursive:
            if path is not '/':
                self.setinfo(path, chmod_cache_info)
            # create a walker
            sub_dir = self.opendir(path)
            for _path, _ in sub_dir.walk.info():
                self.setinfo(path + _path, chmod_cache_info)
            sub_dir.close()
        else:
            self.setinfo(path, chmod_cache_info)

    def mount_fs(self,
                 dst_path: str,
                 fs_url: str = None,
                 owner_uid: Optional[int] = 0,
                 group_gid: Optional[int] = 0,
                 perms: Optional[Union[Permissions, int]] = 0o755) -> subfs.SubFS:
        """
        To be called to mount individual filesystems.
        :param fs_url: Location/URL for the file system that is to be mounted.
        :param dst_path: Place in the Conpot's file system where the files would be placed. This should be relative to
        FS root.
        :param owner_uid: The owner `user` **UID** of the directory and the sub directory. Default is root/
        :param group_gid: The group 'group` to which the directory beings. Defaults to root.
        :param perms: Permission UMASK
        """
        path = self.norm_path(dst_path)
        if self.exists(path) and self.isdir(path):
            if not fs_url:
                new_dir = self.create_jail(path)
            else:
                temp_fs = open_fs(fs_url=fs_url)
                with temp_fs.lock():
                        new_dir = self.opendir(self.norm_path(dst_path), factory=SubAbstractFS)
                        mirror.mirror(src_fs=temp_fs, dst_fs=new_dir)
                        self._cache.update({path: info for path, info in self.walk.info(namespaces=['basic', 'access',
                                                                                                    'details',
                                                                                                    'stat'])})
                del temp_fs  # delete the instance since no longer required
            new_dir.default_uid, new_dir.default_gid = owner_uid, group_gid
            new_dir.chown('/', uid=owner_uid, gid=group_gid, recursive=True)
            new_dir.chmod('/', mode=perms, recursive=True)
            return new_dir
        else:
            raise fs.errors.DirectoryExpected('{} path does not exists'.format(path))

    def __getattribute__(self, attr):
        # Restrict access to methods that are implemented in AbstractFS class - Calling methods from base class may
        # not be safe to use.
        # FIXME: Need to fix these for only allow methods that are defined here.
        if not WrapFS:
            return
        method_list = [x for x, y in WrapFS.__dict__.items() if type(y) == FunctionType]
        if attr in method_list:
            if attr in super(AbstractFS, self).__getattribute__('__dict__').keys() or \
                    attr not in ['match', 'settext']:
                # These methods have been overwritten and are safe to use.
                try:
                    return super(AbstractFS, self).__getattribute__(attr)
                except KeyError as ke:
                    raise FilesystemError('Invalid Path : {}'.format(ke))
            else:
                raise NotImplementedError('The method requested is not supported by Conpot\'s VFS')
        else:
            try:
                return super(AbstractFS, self).__getattribute__(attr)
            except KeyError as ke:
                raise FilesystemError('Invalid Path : {}'.format(ke))


# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import socket

import hpfeeds
import gevent
import logging

logger = logging.getLogger(__name__)


class HPFriendsLogger(object):

    def __init__(self, host, port, ident, secret, channels):
        self.host = host
        self.port = port
        self.ident = ident
        self.secret = secret
        self.channels = channels
        self.max_retires = 5
        self._initial_connection_happend = False
        self.greenlet = gevent.spawn(self._start_connection, host, port, ident, secret)

    def _start_connection(self, host, port, ident, secret):
        # if no initial connection to hpfeeds this will hang forever, reconnect=True only comes into play
        # when lost connection after the initial connect happend.
        self.hpc = hpfeeds.new(host, port, ident, secret)
        self._initial_connection_happend = True

    def log(self, data):
        retries = 0
        if self._initial_connection_happend:
            # hpfeed lib supports passing list of channels
            while True:
                if retries >= self.max_retires:
                    break
                try:
                    self.hpc.publish(self.channels, data)
                except socket.error:
                    retries += 1
                    self.__init__(self.host, self.port, self.ident, self.secret, self.channels)
                    gevent.sleep(0.5)
                else:
                    break
            error_msg = self.hpc.wait()
            return error_msg
        else:
            error_msg = 'Not logging event because initial hpfeeds connect has not happend yet.'
            logger.warning(error_msg)
            return error_msg





from modbus_tk.hooks import call_hooks
import conpot.core as conpot_core


class ModbusBlockDatabusMediator(object):
    """This class represents the values for a range of addresses"""

    def __init__(self, databus_key, starting_address):
        """
        Constructor: defines the address range and creates the array of values
        """
        self.starting_address = starting_address
        # self._data = [0]*size
        self.databus_key = databus_key
        self.size = len(conpot_core.get_databus().get_value(self.databus_key))

    def is_in(self, starting_address, size):
        """
        Returns true if a block with the given address and size
        would overlap this block
        """
        if starting_address > self.starting_address:
            return (self.starting_address+self.size) > starting_address
        elif starting_address < self.starting_address:
            return (starting_address + size) > self.starting_address
        return True

    def __getitem__(self, r):
        """"""
        return conpot_core.get_databus().get_value(self.databus_key).__getitem__(r)

    def __setitem__(self, r, v):
        """"""
        call_hooks("modbus.ModbusBlock.setitem", (self, r, v))
        obj = conpot_core.get_databus().get_value(self.databus_key)
        return obj.__setitem__(r, v)



# Copyright (C) 2015  Adarsh Dinesh <adarshdinesh@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import subprocess

logger = logging.getLogger(__name__)


def _check_mac(iface, addr):
    s = subprocess.Popen(['ifconfig', iface], stdout=subprocess.PIPE)
    data = s.stdout.read()
    if addr in data:
        return True
    else:
        return False


def _is_dhcp(iface):
    s = subprocess.Popen(['cat', '/var/lib/dhcp/dhclient.leases'], stdout=subprocess.PIPE)
    data = s.stdout.read()
    if iface in data:
        return True
    else:
        return False


def _renew_lease(iface):
    subprocess.Popen(['dhclient', '-r'], stdout=subprocess.PIPE)
    subprocess.Popen(['dhclient', iface], stdout=subprocess.PIPE)


def change_mac(iface=None, mac=None, config=None, revert=None):
    if config:
        iface = config.get('change_mac_addr', 'iface')
        mac = config.get('change_mac_addr', 'addr')

    # Changing MAC address and restarting network
    subprocess.Popen(['ip', 'link', 'set', iface, 'down'], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    subprocess.Popen(['ip', 'link', 'set', 'dev', iface, 'address', mac], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    subprocess.Popen(['ip', 'link', 'set', iface, 'up'], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)

    if _check_mac(iface, mac):
        if revert:
            logger.info('MAC address reverted for interface %s', iface)
        else:
            logger.info('MAC address of interface %s changed %s', iface, mac)
        if _is_dhcp(iface):
            _renew_lease(iface)
            logger.info('Interface has a DHCP lease, refreshed.')
    else:
        logger.warning('Could not change MAC address.')


def revert_mac(iface):
    s = subprocess.Popen(['ethtool', '-P', iface], stdout=subprocess.PIPE)
    mac = s.stdout.read().split(' ')[2].strip()
    change_mac(iface, mac, revert=True)



# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import tempfile
import shutil
import os

from lxml import etree
from conpot.core.protocol_wrapper import conpot_protocol
from conpot.protocols.snmp.command_responder import CommandResponder
from conpot.protocols.snmp.build_pysnmp_mib_wrapper import find_mibs, compile_mib
import conpot.core as conpot_core


logger = logging.getLogger()


@conpot_protocol
class SNMPServer(object):
    def __init__(self, template, template_directory, args):
        """
        :param host:        hostname or ip address on which to server the snmp service (string).
        :param port:        listen port (integer).
        :param template:    path to the protocol specific xml configuration file (string).
        """

        self.dom = etree.parse(template)
        self.cmd_responder = None

        if args.mibpaths:
            self.compiled_mibs = args.mibpaths
        else:
            self.compiled_mibs = [os.path.join(template_directory, 'snmp', 'mibs_compiled')]

        if args.raw_mib:
            self.raw_mibs = args.raw_mib
        else:
            self.raw_mibs = [os.path.join(template_directory, 'snmp', 'mibs_raw')]

    def xml_general_config(self, dom):
        snmp_config = dom.xpath('//snmp/config/*')
        if snmp_config:
            for entity in snmp_config:

                # TARPIT: individual response delays
                if entity.attrib['name'].lower() == 'tarpit':

                    if entity.attrib['command'].lower() == 'get':
                        self.cmd_responder.resp_app_get.tarpit = self.config_sanitize_tarpit(entity.text)
                    elif entity.attrib['command'].lower() == 'set':
                        self.cmd_responder.resp_app_set.tarpit = self.config_sanitize_tarpit(entity.text)
                    elif entity.attrib['command'].lower() == 'next':
                        self.cmd_responder.resp_app_next.tarpit = self.config_sanitize_tarpit(entity.text)
                    elif entity.attrib['command'].lower() == 'bulk':
                        self.cmd_responder.resp_app_bulk.tarpit = self.config_sanitize_tarpit(entity.text)

                # EVASION: response thresholds
                if entity.attrib['name'].lower() == 'evasion':

                    if entity.attrib['command'].lower() == 'get':
                        self.cmd_responder.resp_app_get.threshold = self.config_sanitize_threshold(entity.text)
                    elif entity.attrib['command'].lower() == 'set':
                        self.cmd_responder.resp_app_set.threshold = self.config_sanitize_threshold(entity.text)
                    elif entity.attrib['command'].lower() == 'next':
                        self.cmd_responder.resp_app_next.threshold = self.config_sanitize_threshold(entity.text)
                    elif entity.attrib['command'].lower() == 'bulk':
                        self.cmd_responder.resp_app_bulk.threshold = self.config_sanitize_threshold(entity.text)

    def xml_mib_config(self, dom, mibpaths, rawmibs_dirs):
        try:
            mibs = dom.xpath('//snmp/mibs/*')
            tmp_mib_dir = tempfile.mkdtemp()
            mibpaths.append(tmp_mib_dir)
            available_mibs = find_mibs(rawmibs_dirs)

            databus = conpot_core.get_databus()
            # parse mibs and oid tables
            for mib in mibs:
                mib_name = mib.attrib['name']
                # compile the mib file if it is found and not already loaded.
                if mib_name in available_mibs and not self.cmd_responder.has_mib(mib_name):
                    compile_mib(mib_name, tmp_mib_dir)
                for symbol in mib:
                    symbol_name = symbol.attrib['name']

                    # retrieve instance from template
                    if 'instance' in symbol.attrib:
                        # convert instance to (int-)tuple
                        symbol_instance = symbol.attrib['instance'].split('.')
                        symbol_instance = tuple(map(int, symbol_instance))
                    else:
                        # use default instance (0)
                        symbol_instance = (0,)


                    # retrieve value from databus
                    value = databus.get_value(symbol.xpath('./value/text()')[0])
                    profile_map_name = symbol.xpath('./value/text()')[0]

                    # register this MIB instance to the command responder
                    self.cmd_responder.register(mib_name,
                                                symbol_name,
                                                symbol_instance,
                                                value,
                                                profile_map_name)
        finally:
            # cleanup compiled mib files
            shutil.rmtree(tmp_mib_dir)

    def config_sanitize_tarpit(self, value):

        # checks tarpit value for being either a single int or float,
        # or a series of two concatenated integers and/or floats separated by semicolon and returns
        # either the (sanitized) value or zero.

        if value is not None:

            x, _, y = value.partition(';')

            try:
                _ = float(x)
            except ValueError:
                logger.error("SNMP invalid tarpit value: '%s'. Assuming no latency.", value)
                # first value is invalid, ignore the whole setting.
                return '0;0'

            try:
                _ = float(y)
                # both values are fine.
                return value
            except ValueError:
                # second value is invalid, use the first one.
                return x

        else:
            return '0;0'

    def config_sanitize_threshold(self, value):

        # checks DoS thresholds for being either a single int or a series of two concatenated integers
        # separated by semicolon and returns either the (sanitized) value or zero.

        if value is not None:

            x, _, y = value.partition(';')

            try:
                _ = int(x)
            except ValueError:
                logger.error("SNMP invalid evasion threshold: '%s'. Assuming no DoS evasion.", value)
                # first value is invalid, ignore the whole setting.
                return '0;0'

            try:
                _ = int(y)
                # both values are fine.
                return value
            except ValueError:
                # second value is invalid, use the first and ignore the second.
                return str(x) + ';0'

        else:
            return '0;0'

    def start(self, host, port):
        self.cmd_responder = CommandResponder(host, port, self.compiled_mibs)
        self.xml_general_config(self.dom)
        self.xml_mib_config(self.dom, self.compiled_mibs, self.raw_mibs)

        logger.info('SNMP server started on: %s', (host, self.get_port()))
        self.cmd_responder.serve_forever()

    def stop(self):
        if self.cmd_responder:
            self.cmd_responder.stop()

    def get_port(self):
        if self.cmd_responder:
            return self.cmd_responder.server_port
        else:
            return None



# Command Responder (GET/GETNEXT)
# Based on examples from http://pysnmp.sourceforge.net/

import logging

from pysnmp.entity import config
from pysnmp.entity.rfc3413 import context
from pysnmp.carrier.asynsock.dgram import udp
from pysnmp.entity import engine
from pysnmp.smi import builder
import gevent

from conpot.protocols.snmp import conpot_cmdrsp
from conpot.protocols.snmp.databus_mediator import DatabusMediator
from gevent.server import DatagramServer

logger = logging.getLogger(__name__)


class SNMPDispatcher(DatagramServer):
    def __init__(self):
        self.__timerResolution = 0.5

    def registerRecvCbFun(self, recvCbFun, recvId=None):
        self.recvCbFun = recvCbFun

    def handle(self, msg, address):
        try:
            self.recvCbFun(self, self.transportDomain, address, msg)
        except Exception as e:
            logger.info("SNMP Exception: %s", e)

    def registerTransport(self, tDomain, transport):
        DatagramServer.__init__(self, transport, self.handle)
        self.transportDomain = tDomain

    def registerTimerCbFun(self, timerCbFun, tickInterval=None):
        pass

    def sendMessage(self, outgoingMessage, transportDomain, transportAddress):
        self.socket.sendto(outgoingMessage, transportAddress)

    def getTimerResolution(self):
        return self.__timerResolution


class CommandResponder(object):
    def __init__(self, host, port, mibpaths):

        self.oid_mapping = {}
        self.databus_mediator = DatabusMediator(self.oid_mapping)
        # mapping between OID and databus keys

        # Create SNMP engine
        self.snmpEngine = engine.SnmpEngine()

        # path to custom mibs
        mibBuilder = self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder
        mibSources = mibBuilder.getMibSources()

        for mibpath in mibpaths:
            mibSources += (builder.DirMibSource(mibpath),)
        mibBuilder.setMibSources(*mibSources)

        # Transport setup
        udp_sock = gevent.socket.socket(gevent.socket.AF_INET, gevent.socket.SOCK_DGRAM)
        udp_sock.setsockopt(gevent.socket.SOL_SOCKET, gevent.socket.SO_BROADCAST, 1)
        udp_sock.bind((host, port))
        self.server_port = udp_sock.getsockname()[1]
        # UDP over IPv4
        self.addSocketTransport(
            self.snmpEngine,
            udp.domainName,
            udp_sock
        )

        # SNMPv1
        config.addV1System(self.snmpEngine, 'public-read', 'public')

        # SNMPv3/USM setup
        # user: usr-md5-des, auth: MD5, priv DES
        config.addV3User(
            self.snmpEngine, 'usr-md5-des',
            config.usmHMACMD5AuthProtocol, 'authkey1',
            config.usmDESPrivProtocol, 'privkey1'
        )
        # user: usr-sha-none, auth: SHA, priv NONE
        config.addV3User(
            self.snmpEngine, 'usr-sha-none',
            config.usmHMACSHAAuthProtocol, 'authkey1'
        )
        # user: usr-sha-aes128, auth: SHA, priv AES/128
        config.addV3User(
            self.snmpEngine, 'usr-sha-aes128',
            config.usmHMACSHAAuthProtocol, 'authkey1',
            config.usmAesCfb128Protocol, 'privkey1'
        )

        # Allow full MIB access for each user at VACM
        config.addVacmUser(self.snmpEngine, 1, 'public-read', 'noAuthNoPriv',
                           readSubTree=(1, 3, 6, 1, 2, 1), writeSubTree=(1, 3, 6, 1, 2, 1))
        config.addVacmUser(self.snmpEngine, 2, 'public-read', 'noAuthNoPriv',
                           readSubTree=(1, 3, 6, 1, 2, 1), writeSubTree=(1, 3, 6, 1, 2, 1))
        config.addVacmUser(self.snmpEngine, 3, 'usr-md5-des', 'authPriv',
                           readSubTree=(1, 3, 6, 1, 2, 1), writeSubTree=(1, 3, 6, 1, 2, 1))
        config.addVacmUser(self.snmpEngine, 3, 'usr-sha-none', 'authNoPriv',
                           readSubTree=(1, 3, 6, 1, 2, 1), writeSubTree=(1, 3, 6, 1, 2, 1))
        config.addVacmUser(self.snmpEngine, 3, 'usr-sha-aes128', 'authPriv',
                           readSubTree=(1, 3, 6, 1, 2, 1), writeSubTree=(1, 3, 6, 1, 2, 1))

        # Get default SNMP context this SNMP engine serves
        snmpContext = context.SnmpContext(self.snmpEngine)

        # Register SNMP Applications at the SNMP engine for particular SNMP context
        self.resp_app_get = conpot_cmdrsp.c_GetCommandResponder(self.snmpEngine, snmpContext, self.databus_mediator)
        self.resp_app_set = conpot_cmdrsp.c_SetCommandResponder(self.snmpEngine, snmpContext, self.databus_mediator)
        self.resp_app_next = conpot_cmdrsp.c_NextCommandResponder(self.snmpEngine, snmpContext, self.databus_mediator)
        self.resp_app_bulk = conpot_cmdrsp.c_BulkCommandResponder(self.snmpEngine, snmpContext, self.databus_mediator)

    def addSocketTransport(self, snmpEngine, transportDomain, transport):
        """Add transport object to socket dispatcher of snmpEngine"""
        if not snmpEngine.transportDispatcher:
            snmpEngine.registerTransportDispatcher(SNMPDispatcher())
        snmpEngine.transportDispatcher.registerTransport(transportDomain, transport)

    def register(self, mibname, symbolname, instance, value, profile_map_name):
        """Register OID"""
        self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.loadModules(mibname)
        s = self._get_mibSymbol(mibname, symbolname)

        if s:
            self.oid_mapping[s.name+instance] = profile_map_name

            MibScalarInstance, = self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMPv2-SMI',
                                                                                                            'MibScalarInstance')
            x = MibScalarInstance(s.name, instance, s.syntax.clone(value))
            self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.exportSymbols(mibname, x)

            logger.debug('Registered: OID %s Instance %s ASN.1 (%s @ %s) value %s dynrsp.', s.name, instance, s.label, mibname, value)

        else:
            logger.debug('Skipped: OID for symbol %s not found in MIB %s', symbolname, mibname)

    def _get_mibSymbol(self, mibname, symbolname):
        modules = self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.mibSymbols
        if mibname in modules:
            if symbolname in modules[mibname]:
                return modules[mibname][symbolname]

    def has_mib(self, mibname):
        modules = self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.mibSymbols
        return mibname in modules

    def serve_forever(self):
        self.snmpEngine.transportDispatcher.serve_forever()

    def stop(self):
        self.snmpEngine.transportDispatcher.stop_accepting()


if __name__ == "__main__":
    server = CommandResponder()
    print('Starting echo server on port 161')
    server.serve_forever()



# Copyright (C) 2013  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging

import libtaxii
import libtaxii.messages
from libtaxii.messages_11 import ContentBlock, InboxMessage, generate_message_id
from libtaxii.clients import HttpClient

from conpot.core.loggers.stix_transform import StixTransformer

logger = logging.getLogger(__name__)


class TaxiiLogger(object):
    def __init__(self, config, dom):
        self.host = config.get('taxii', 'host')
        self.port = config.getint('taxii', 'port')
        self.inbox_path = config.get('taxii', 'inbox_path')
        self.use_https = config.getboolean('taxii', 'use_https')

        self.client = HttpClient()
        self.client.setProxy('noproxy')
        self.stix_transformer = StixTransformer(config, dom)

    def log(self, event):
        # converts from conpot log format to STIX compatible xml
        stix_package = self.stix_transformer.transform(event)

        # wrapping the stix message in a TAXII envelope
        content_block = ContentBlock(libtaxii.CB_STIX_XML_11, stix_package.encode('utf-8'))
        inbox_message = InboxMessage(message_id=generate_message_id(), content_blocks=[content_block])
        inbox_xml = inbox_message.to_xml()

        # the actual call to the TAXII web service
        response = self.client.callTaxiiService2(self.host, self.inbox_path, libtaxii.VID_TAXII_XML_11, inbox_xml, self.port)
        response_message = libtaxii.get_message_from_http_response(response, '0')

        if response_message.status_type != libtaxii.messages.ST_SUCCESS:
            logger.error('Error while transmitting message to TAXII server: %s', response_message.message)
            return False
        else:
            return True




# This module is based on the original work done by Giampaolo Rodola and pyftpdlib authors.
# This is a heavily customized version that supports Conpot's virtual file system os.* wrappers and gevent support.

from conpot.protocols.ftp.ftp_base_handler import FTPHandlerBase
import logging
import fs
import os
import glob
import sys
import tempfile
from datetime import datetime
import gevent
from gevent import socket
from fs import errors
from conpot.core.filesystem import FilesystemError, FSOperationNotPermitted
from conpot.protocols.ftp.ftp_utils import FTPPrivilegeException, get_data_from_iter
logger = logging.getLogger(__name__)

# -----------------------------------------------------
# *Implementation Note*: Regarding Permissions:
# -----------------------------------------------------
# To change a directory as current directory we need permissions : rwx (CWD)| Also it has to be a directory. |
# To read a file we need : r permissions - It has to be a file | list files (LIST, NLST, STAT, SIZE, MDTM
# RETR commands)
# To store a file to the server we need 'w' permissions - (STOR, STOU commands)
# To rename file or directory we need 'w' permissions (RNFR, RNTO)
# To delete file or directory we need 'w' permissions (DELE, RMD commands)
# To append data to an existing file (APPE command) we need 'w' permissions
# -----------------------------------------------------


class FTPCommandChannel(FTPHandlerBase):
    """
    FTP Command Responder. Implementation of RFC 959.
    """

    # -----------------------------------------------------------------------
    # There are some commands that do not require any kind of auth and permissions to run.
    # These also do not require us to have an established data channel. So let us get rid of those first.
    # Btw: These are - USER, PASS, HELP, NOOP, SYST, QUIT, SITE HELP, PWD, TYPE
    # All commands assume that the path supplied have been cleaned and sanitized.

    # The USER command is used to verify users as they try to login.
    def do_USER(self, arg):
        """
        USER FTP command. If the user is already logged in, return 530 else 331 for the PASS command
        :param arg: username specified by the client/attacker
        """
        # first we need to check if the user is authenticated?
        if self.authenticated:
            self.respond(b'530 Cannot switch to another user.')
        else:
            self.username = arg
            self.respond(b'331 Now specify the Password.')

    def do_PASS(self, arg):
        if self.authenticated:
            self.respond(b"503 User already authenticated.")
        if not self.username:
            self.respond(b"503 Login with USER first.")
        if self.authentication_ok(user_pass=arg):
            if not self.config.motd:
                self.respond(b'230 Log in Successful.')
            else:
                _msg = '220-{}\r\n'.format(self.config.motd)
                self.respond(_msg)
                self.respond(b'220 ')
        else:
            self.invalid_login_attempt += 1
            self.respond(b'530 Authentication Failed.')

    def do_HELP(self, arg):
        """Return help text to the client."""
        if arg:
            line = arg.upper()
            if line in self.config.COMMANDS:
                self.respond(b'214 %a' % self.config.COMMANDS[line]['help'])
            else:
                self.respond(b'501 Unrecognized command.')
        else:
            # provide a compact list of recognized commands
            def formatted_help():
                cmds = []
                keys = sorted([x for x in self.config.COMMANDS.keys()
                               if not x.startswith('SITE ')])
                while keys:
                    elems = tuple((keys[0:8]))
                    cmds.append(b' %-6a' * len(elems) % elems + b'\r\n')
                    del keys[0:8]
                return b''.join(cmds)

            _buffer = b'214-The following commands are recognized:\r\n'
            _buffer += formatted_help()
            self.respond(_buffer + b'214 Help command successful.')

    def do_NOOP(self, arg):
        """Do nothing. No params required. No auth required and no permissions required."""
        self.respond(b'200 I successfully done nothin\'.')

    def do_SYST(self, arg):
        """Return system type (always returns UNIX type: L8)."""
        # This command is used to find out the type of operating system
        # at the server.  The reply shall have as its first word one of
        # the system names listed in RFC-943.
        # Since that we always return a "/bin/ls -lA"-like output on
        # LIST we  prefer to respond as if we would on Unix in any case.
        self.respond(b'215 UNIX Type: L8')

    def do_QUIT(self, arg):
        self.respond(b'221 Bye.')
        self.session.add_event({'type': 'CONNECTION_TERMINATED'})
        self.disconnect_client = True

    def do_SITE_HELP(self, line):
        """Return help text to the client for a given SITE command."""
        if line:
            line = line.upper()
            if line in self.config.COMMANDS:
                self.respond(b'214 %a' % self.config.COMMANDS[line]['help'])
            else:
                self.respond(b'501 Unrecognized SITE command.')
        else:
            _buffer = b'214-The following SITE commands are recognized:\r\n'
            site_cmds = []
            for cmd in sorted(self.config.COMMANDS.keys()):
                if cmd.startswith('SITE '):
                    site_cmds.append(b' %a\r\n' % cmd[5:])
            _buffer_cmds = b''.join(site_cmds)
            self.respond(_buffer + _buffer_cmds + b'214 Help SITE command successful.')

    def do_MODE(self, line):
        """Set data transfer mode ("S" is the only one supported (noop))."""
        mode = line.upper()
        if mode == 'S':
            self.respond(b'200 Transfer mode set to: S')
        elif mode in ('B', 'C'):
            self.respond(b'504 Unimplemented MODE type.')
        else:
            self.respond(b'501 Unrecognized MODE type.')

    def do_PWD(self, arg):
        """Return the name of the current working directory to the client."""
        pwd = self.working_dir
        try:
            assert isinstance(pwd, str), pwd
            _pwd = '257 "{}" is the current directory.'.format(pwd)
            self.respond(_pwd.encode())
        except AssertionError:
            logger.info('FTP CWD specified is not unicode. {}'.format(pwd))
            self.respond(b'FTP CWD not unicode.')

    def do_TYPE(self, line):
        """Set current type data type to binary/ascii"""
        data_type = line.upper().replace(' ', '')
        if data_type in ("A", "L7"):
            self.respond(b'200 Type set to: ASCII.')
            self._current_type = 'a'
        elif data_type in ("I", "L8"):
            self.respond(b'200 Type set to: Binary.')
            self._current_type = 'i'
        else:
            self.respond(b'504 Unsupported type "%a".' % line)

    def do_STRU(self, line):
        """Set file structure ("F" is the only one supported (noop))."""
        stru = line.upper()
        if stru == 'F':
            self.respond(b'200 File transfer structure set to: F.')
        elif stru in ('P', 'R'):
            self.respond(b'504 Unimplemented STRU type.')
        else:
            self.respond(b'501 Unrecognized STRU type.')

    def do_ALLO(self, arg):
        """Allocate bytes for storage (noop)."""
        # not necessary (always respond with 202)
        self.respond(b'202 No storage allocation necessary.')

    def do_REIN(self, arg):
        """Reinitialize user's current session."""
        self.stop_data_channel()
        self.username = None
        self._uid = None
        self.authenticated = False
        self.respond(b'230 Ready for new user.')

    # -----------------------------------------------------------------------
    # Next up we have commands that may require some kind of auth and permissions to run.
    # These also do not require us to have an established data channel.
    # These commands are -
    #  - {MDTM, SIZE, STAT, DELE, RNFR, RNTO} - require read permissions
    #  - {MKD, RMD} - require write permissions + path should be files
    #  - {CDUP, CWD and CHMOD} - require 'rwx' permissions + path should be folders
    # Again all commands assume that the path supplied have been cleaned and sanitized.

    def do_MDTM(self, path):
        """Return last modification time of file to the client as an ISO
        3307 style timestamp (YYYYMMDDHHMMSS) as defined in RFC-3659.
        On success return the file path, else None.
        """
        try:
            path = self.ftp_path(path)
            if not self.config.vfs.isfile(path):
                _msg = '550 {} is not retrievable'.format(path)
                self.respond(_msg.encode())
                return
            with self.config.vfs.check_access(path=path, user=self._uid, perms='r'):
                m_time = '213 {}'.format(self.config.vfs.getmtime(path).strftime("%Y%m%d%H%M%S"))
                self.respond(m_time.encode())
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (ValueError, fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            # It could happen if file's last modification time
            # happens to be too old (prior to year 1900)
            self.respond(b'550 Can\'t determine file\'s last modification time.')

    def do_SIZE(self, path):
        """Return size of file in a format suitable for using with RESTart as defined in RFC-3659."""
        try:
            path = self.ftp_path(path)
            if self._current_type == 'a':
                self.respond(b'550 SIZE not allowed in ASCII mode.')
                return
            # If the file is a sym-link i.e. not readable, send not retrievable
            if not self.config.vfs.isfile(path):
                self.respond(b'550 is not retrievable.')
                return
            else:
                with self.config.vfs.check_access(path=path, user=self._uid, perms='r'):
                    size = self.config.vfs.getsize(path)
                    self.respond(b'213 %a' % size)
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (OSError, fs.errors.FSError) as err:
            self.respond(b'550 %a.' % self._log_err(err))

    def do_STAT(self, path):
        """If invoked without parameters, returns general status information about the FTP server process.
        If a parameter is given, acts like the LIST command, except that data is sent over the command
        channel (no PORT or PASV command is required).
        """
        # return STATus information about ftp data connection
        if not path:
            s = list()
            s.append('Connected to: {}:{}'.format(self.host, self.port))
            if self.authenticated:
                s.append('Logged in as: {}'.format(self.username))
            else:
                if not self.username:
                    s.append("Waiting for username.")
                else:
                    s.append("Waiting for password.")
            if self._current_type == 'a':
                _type = 'ASCII'
            else:
                _type = 'Binary'
            s.append("TYPE: {}; STRUcture: File; MODE: Stream".format(_type))
            if self._data_sock is not None and self._data_channel is False:
                s.append('Passive data channel waiting for connection.')
            elif self._data_channel is True:
                bytes_sent = self.metrics.data_channel_bytes_send + self.metrics.command_chanel_bytes_send
                bytes_recv = self.metrics.command_chanel_bytes_recv + self.metrics.data_channel_bytes_recv
                elapsed_time = self.metrics.get_elapsed_time()
                s.append('Data connection open:')
                s.append('Total bytes sent: {}'.format(bytes_sent))
                s.append('Total bytes received: {}'.format(bytes_recv))
                s.append('Transfer elapsed time: {} secs'.format(elapsed_time))
            else:
                s.append('Data connection closed.')

            self.respond('211-FTP server status:\r\n')
            self.respond(''.join([' {}\r\n'.format(item) for item in s]))
            self.respond('211 End of status.')
        # return directory LISTing over the command channel
        else:
            try:
                line = self.ftp_path(path)
                with self.config.vfs.check_access(path=line, user=self._uid, perms='r'):
                    if self.config.vfs.isdir(path):
                        listing = self.config.vfs.listdir(path)
                        # RFC 959 recommends the listing to be sorted.
                        listing.sort()
                        iterator = self.config.vfs.format_list(path, listing)
                    else:
                        basedir, filename = os.path.split(path)
                        self.config.stat(path)
                        iterator = self.config.vfs.format_list(basedir, [filename])
                    _status = '213-Status of "{}":\r\n'.format(line)
                    _status += get_data_from_iter(iterator)
                    _status += '213 End of status.'
                    self.respond(_status.encode())
            except FSOperationNotPermitted:
                self.respond(b'500 Operation not permitted.')
            except (OSError, FilesystemError, AssertionError, fs.errors.FSError, FTPPrivilegeException):
                self.respond(b'550 STAT command failed.')

    def do_MKD(self, path):
        """
        Create the specified directory. On success return the directory path, else None.
        """
        try:
            # In order to create a directory the current user must have 'w' permissions for the parent directory
            # of current path.
            _dir = self.ftp_path(path)
            with self.config.vfs.check_access(path=self.working_dir, user=self._uid, perms='w'):
                self.config.vfs.makedir(_dir)
            _mkd = '257 "{}" directory created.'.format(_dir)
            self.respond(_mkd)
            self.config.vfs.chmod(_dir, self.config.dir_default_perms)
            self.config.vfs.chown(_dir, uid=self._uid, gid=self.config.get_gid(self._uid))
            self.config.vfs.settimes(_dir, datetime.now(), datetime.now())
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (FilesystemError, fs.errors.FSError, FTPPrivilegeException):
            self.respond(b'550 Create directory operation failed.')

    def do_RMD(self, path):
        """Remove the specified directory. On success return the directory path, else None.
        """
        if self.ftp_path(path) == self.working_dir or path == '/':
            self.respond(b'550 Can\'t remove root directory.')
            return
        try:
            _path = self.ftp_path(self.working_dir + path)
            # In order to create a directory the current user must have 'w' permissions for the current directory
            with self.config.vfs.check_access(path=self.ftp_path(os.path.join(path, '../')),
                                              user=self._uid, perms='w'):
                self.config.vfs.removedir(_path)
            self.respond(b'250 Directory removed.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 Remove directory operation failed.')

    def do_CWD(self, path):
        """Change the current working directory."""
        # Temporarily join the specified directory to see if we have permissions to do so, then get back to original
        # process's current working directory.
        try:
            init_cwd = self.working_dir
            if not self.config.vfs.isdir(path):
                raise FSOperationNotPermitted
            # make sure the current user has permissions to the new dir. To change the directory, user needs to have
            # executable permissions for the directory
            with self.config.vfs.check_access(path=path, user=self._uid, perms='rwx'):
                self.working_dir = self.ftp_path(path)
            logger.info('Changing current directory {} to {}'.format(init_cwd, self.working_dir))
            _cwd = '250 "{}" is the current directory.'.format(self.working_dir)
            self.respond(_cwd.encode())
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 Failed to change directory.')

    def do_CDUP(self, arg):
        """
        Change into the parent directory. On success return the new directory, else None.
        """
        # Note: RFC-959 says that code 200 is required but it also says
        # that CDUP uses the same codes as CWD.
        return self.do_CWD(path='/'.join([self.ftp_path(self.working_dir), '../']))

    def do_DELE(self, path):
        """Delete the specified file."""
        try:
            path = self.ftp_path(path)
            if not self.config.vfs.isfile(path):
                self.respond(b'550 Failed to delete file.')
            else:
                with self.config.vfs.check_access(path=path, user=self._uid, perms='w'):
                    self.config.vfs.remove(path)
                    self.respond(b'250 File removed.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 Failed to delete file.')

    def do_RNFR(self, path):
        """Rename the specified (only the source name is specified
        here, see RNTO command)"""
        try:
            path = self.ftp_path(path)
            if self.config.vfs.isfile(path) or self.config.vfs.isdir(path):
                with self.config.vfs.check_access(path=path, user=self._uid, perms='w'):
                    assert isinstance(path, str)
                    if path == '/':
                        self.respond(b"550 Can't rename home directory.")
                    else:
                        self._rnfr = path
                        self.respond(b'350 Ready for destination name.')
            else:
                # file neither a file or a directory.
                raise AssertionError
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (AssertionError, KeyError, fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 No such file or directory.')

    def do_RNTO(self, dst_path):
        """Rename file (destination name only, source is specified with RNFR)."""
        try:
            assert isinstance(dst_path, str)
            if not self._rnfr:
                self.respond(b'503 Bad sequence of commands: use RNFR first.')
                return
            src = self.ftp_path(self._rnfr)
            self._rnfr = None
            if self.config.vfs.isdir(src):
                _move = self.config.vfs.movedir
            elif self.config.vfs.isfile(src):
                _move = self.config.vfs.move
            else:
                raise FilesystemError
            with self.config.vfs.check_access(path=src, user=self._uid, perms='w'):
                _path, _file = os.path.split(src)
                _, _dst_file = os.path.split(dst_path)
                # create new paths
                _file = os.path.join(_path, _file)
                _dst_file = os.path.join(_path, _dst_file)
                if _file != _dst_file:
                    logger.info('Renaming file from {} to {}'.format(_file, _dst_file))
                    _move(_file, _dst_file, overwrite=True)
                self.respond(b"250 Renaming ok.")
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (ValueError, fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 File rename operation failed.')

    def do_SITE_CHMOD(self, path, mode):
        """Change file mode. On success return a (file_path, mode) tuple."""
        try:
            # Note: although most UNIX servers implement it, SITE CHMOD is not
            # defined in any official RFC.
            path = self.ftp_path(path)
            with self.config.vfs.check_access(path=path, user=self._uid, perms='rwx'):
                assert len(mode) in (3, 4)
                for x in mode:
                    assert 0 <= int(x) <= 7
                mode = int(mode, 8)
                # To do a chmod user needs to be the owner of the file.
                self.config.vfs.chmod(path, mode)
                self.respond(b'200 SITE CHMOD successful.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (AssertionError, ValueError):
            self.respond(b"501 Invalid SITE CHMOD format.")
        except (fs.errors.FSError, FilesystemError, FTPPrivilegeException):
            self.respond(b'550 SITE CHMOD command failed.')

    # -----------------------------------------------------------------------
    # Following up we have the PORT(active) and PASV(passive) commnads.
    # These setup the data channel for data transfer.

    def do_PASV(self, arg):
        """
            Starts a Passive Data Channel using IPv4. We don't actually need to start the full duplex connection here.
            Just need to figure the host ip and the port. The DTP connection would start in each command.
        """
        if self._data_channel:
            self.stop_data_channel(purge=True, reason='Switching from PASV mode.')
        self.active_passive_mode = 'PASV'
        # We are in passive mode. Here we would create a simple socket listener.
        self._data_listener_sock = gevent.socket.socket()
        self._data_listener_sock.bind((self._local_ip, 0))
        ip, port = self._data_listener_sock.getsockname()
        self.respond('227 Entering Passive Mode (%s,%u,%u).' % (','.join(ip.split('.')), port >> 8 & 0xFF,
                                                                port & 0xFF))
        try:
            self._data_listener_sock.listen(1)
            self._data_listener_sock.settimeout(5)  # Timeout for ftp client to send info
            logger.info('Client {} entering FTP passive mode'.format(self.client_address))
            self._data_sock, (self.cli_ip, self.cli_port) = self._data_listener_sock.accept()
            logger.info('Client {} provided ({}:{}) for PASV connection.'.format(
                self.client_address, self.cli_ip, self.cli_port)
            )
            logger.info('FTP: starting data channel for client {}'.format(self.client_address))
            self._data_listener_sock.close()
        except (socket.error, socket.timeout) as se:
            logger.info('Can\'t switch to PASV mode.  Error occurred: {}'.format(str(se)))
            self.respond(b'550 PASV command failed.')

    def do_PORT(self, arg):
        """
        Starts an active data channel by using IPv4. We don't actually need to start the full duplex connection here.
        Just need to figure the host ip and the port. The DTP connection would start in each command.
        """
        if self._data_channel:
            self.stop_data_channel(purge=True, reason='Switching from PORT mode.')
        self.active_passive_mode = 'PORT'
        try:
            addr = list(map(int, arg.split(',')))
            if len(addr) != 6:
                raise ValueError
            for x in addr[:4]:
                if not 0 <= x <= 255:
                    raise ValueError
            ip = '%d.%d.%d.%d' % tuple(addr[:4])
            port = (addr[4] * 256) + addr[5]
            if not 0 <= port <= 65535:
                raise ValueError
            self.cli_ip, self.cli_port = ip, port
            self._data_sock = gevent.socket.socket()
            self._data_sock.connect((self.cli_ip, self.cli_port))
            logger.info('Client {} entered FTP active mode'.format(self.client_address))
            logger.info(
                'Client {} provided {}:{} for PORT connection.'.format(self.client_address, self.cli_ip, self.cli_port)
            )
            self.respond(b'200 PORT Command Successful. Consider using PASV.')
            logger.info('FTP: configured data channel for client {}'.format(self.client_address))
        except (ValueError, OverflowError):
            self.respond("501 Invalid PORT format.")
        except socket.error as se:
            if self._data_channel:
                self.stop_data_channel(reason='Can\'t switch to Active(PORT) mode. Error occurred: {}'.format(str(se)))

    # -- Data Channel related commands --

    def do_LIST(self, path):
        try:
            _path = self.ftp_path(path)
            with self.config.vfs.check_access(path=_path, user=self._uid, perms='r'):
                listing = self.config.vfs.listdir(_path)
                if isinstance(listing, list):
                    # RFC 959 recommends the listing to be sorted.
                    listing.sort()
                    iterator = self.config.vfs.format_list(_path, listing)
                    self.respond('150 Here comes the directory listing.')
                    _list_data = get_data_from_iter(iterator)
                    # Push data to the data channel
                    self.push_data(_list_data.encode())
                    # start the command channel
                    self.start_data_channel()
                    self.respond(b'226 Directory send OK.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (OSError, fs.errors.FSError, FilesystemError, FTPPrivilegeException) as err:
            self._log_err(err)
            self.respond(b'550 LIST command failed.')

    def do_NLST(self, path):
        """Return a list of files in the specified directory in a compact form to the client."""
        try:
            _path = self.ftp_path(path)
            with self.config.vfs.check_access(path=_path, user=self._uid, perms='r'):
                listing = self.config.vfs.listdir(_path)
                data = ''
                if listing:
                    listing.sort()
                    data = '\r\n'.join(listing) + '\r\n'
                self.respond(b'150 Here comes the directory listing.')
                # Push data to the data channel
                self.push_data(data=data)
                # start the command channel
                self.start_data_channel()
                self.respond(b'226 Directory send OK.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (OSError, fs.errors.FSError, FilesystemError, FTPPrivilegeException) as err:
            self._log_err(err)
            self.respond(b'550 NLST command failed.')

    def do_RETR(self, arg):
        """
        Fetch and send a file.
        :param arg: Filename that is to be retrieved
        """
        try:
            filename = self.ftp_path(arg)
            with self.config.vfs.check_access(path=filename, user=self._uid, perms='r'):
                if self.config.vfs.isfile(filename):
                    self.send_file(file_name=filename)
                else:
                    raise FilesystemError('cmd: RETR. Path requested {} is not a file.')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except (OSError, fs.errors.FSError, FilesystemError, FTPPrivilegeException) as err:
            self._log_err(err)
            self.respond(b'550 The system cannot find the file specified.')

    def do_ABOR(self, arg):
        """Aborts a file transfer currently in progress."""
        if self.active_passive_mode is None:
            self.respond(b'225 No transfer to abort.')
        else:
            # a PASV or PORT was received but connection wasn't made yet
            if not self._data_channel:
                self.stop_data_channel(abort=True, purge=True, reason='ABOR called.')
                self.respond(b'225 ABOR command successful; data channel closed.')
            else:
                self.stop_data_channel(abort=True, purge=True, reason='ABOR called.')
                self.respond(b'426 Transfer aborted via ABOR.')
                self.respond(b'226 ABOR command successful.')

    def do_STOR(self, file, mode='w'):
        """Store a file (transfer from the client to the server)."""
        # A resume could occur in case of APPE or REST commands.
        # In that case we have to open file object in different ways:
        # STOR: mode = 'w'
        # APPE: mode = 'a'
        # REST: mode = 'r+' (to permit seeking on file object)
        if 'a' in mode:
            cmd = 'APPE'
        else:
            cmd = 'STOR'
        try:
            with self.config.vfs.check_access(path=self.working_dir, user=self._uid, perms='w'):
                rest_pos = self._restart_position
                self._restart_position = 0
                if rest_pos:
                    if rest_pos > self.config.vfs.getsize(file):
                        raise ValueError('Can\'t seek file more than its size.')
                    # rest_pos != 0 and not None. Must be REST cmd
                    cmd = 'REST'
                else:
                    rest_pos = 0
                if cmd == 'APPE':
                    _file_seek = self.config.vfs.getsize(file)
                elif cmd == 'REST':
                    _file_seek = rest_pos
                else:
                    assert cmd == 'STOR'
                    _file_seek = 0
                self.recv_file(os.path.join(self.working_dir, file), _file_seek, cmd=cmd)
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')
        except ValueError as err:
            self._log_err(err)
            self.respond(b'550 STOR command failed. Can\'t seek file more than its size.')
        except (OSError, AssertionError, fs.errors.FSError, FilesystemError, FTPPrivilegeException) as err:
            self._log_err(err)
            self.respond(b'550 STOR command failed. .')

    def do_REST(self, line):
        """Restart a file transfer from a previous mark."""
        if self._current_type == 'a':
            self.respond(b'501 Resuming transfers not allowed in ASCII mode.')
            return
        try:
            marker = int(line)
            if marker < 0:
                raise ValueError
            else:
                self.respond("350 Restarting at position {}.".format(marker))
                self._restart_position = marker
        except (ValueError, OverflowError):
            self.respond(b"501 Invalid parameter.")

    def do_APPE(self, file):
        """Append data to an existing file on the server.
        On success return the file path, else None.
        """
        # watch for APPE preceded by REST, which makes no sense.
        if self._restart_position:
            self.respond(b"450 Can't APPE while REST request is pending.")
        else:
            return self.do_STOR(file, mode='a')

    def do_STOU(self, line):
        """Store a file on the server with a unique name."""
        try:
            if self._restart_position:
                self.respond(b'450 Can\'t STOU while REST request is pending.')
                return
            _, _file_name = os.path.split(tempfile.NamedTemporaryFile().name)
            if line:
                line = self.ftp_path(line)
                basedir, prefix = os.path.split(line)
                _file_name = '.' + _file_name
            else:
                basedir = self.working_dir
                if self.config.stou_suffix:
                    _file_name = _file_name + self.config.stou_suffix
                if self.config.stou_prefix:
                    _file_name = self.config.stou_prefix + _file_name
            with self.config.vfs.check_access(path=basedir, user=self._uid, perms='w'):
                self.respond(b'150 FILE: %a' % _file_name)
                self.recv_file(os.path.join(basedir, _file_name), 0, cmd='STOR')
        except FSOperationNotPermitted:
            self.respond(b'500 Operation not permitted.')

    # -----------------------------------------------------------------
    # Depreciated/alias commands
    # RFC-1123 requires that the server treat XCUP, XCWD, XMKD, XPWD and XRMD commands as synonyms for CDUP, CWD, MKD,
    # LIST and RMD. Such commands are obsoleted but some ftp clients (e.g. Windows ftp.exe) still use them.

    # Change to the parent directory. Synonym for CDUP. Deprecated.
    do_XCUP = do_CDUP
    # Change the current working directory. Synonym for CWD. Deprecated.
    do_XCWD = do_CWD
    # Create the specified directory. Synonym for MKD. Deprecated.
    do_XMKD = do_MKD
    # Return the current working directory. Synonym for PWD. Deprecated.
    do_XPWD = do_PWD
    # Remove the specified directory. Synonym for RMD. Deprecated.
    do_XRMD = do_RMD
    # Quit and end the current ftp session. Synonym for QUIT
    do_BYE = do_QUIT

    # -----------------------------------------------------------------
    # Helper methods and Command Processors.
    def _log_err(self, err):
        """
        Log errors and send an unexpected response standard message to the client.
        :param err: Exception object
        :return: 500 msg to be sent to the client.
        """
        logger.info('FTP error occurred. Client: {} error {}'.format(self.client_address, str(err)))

    # clean things, sanity checks and more
    def _pre_process_cmd(self, line, cmd, arg):
        kwargs = {}
        if cmd == "SITE" and arg:
            cmd = "SITE %s" % arg.split(' ')[0].upper()
            arg = line[len(cmd) + 1:]

        logger.info('Received command {} : {} from FTP client {}: {}'.format(cmd, line, self.client_address,
                                                                             self.session.id))
        if cmd not in self.config.COMMANDS:
            if cmd[-4:] in ('ABOR', 'STAT', 'QUIT'):
                cmd = cmd[-4:]
            else:
                self.respond(b'500 Command %a not understood' % cmd)
                return

        # - checking for valid arguments
        if not arg and self.config.COMMANDS[cmd]['arg'] is True:
            self.respond(b"501 Syntax error: command needs an argument")
            return
        if arg and self.config.COMMANDS[cmd]['arg'] is False:
            self.respond(b'501 Syntax error: command does not accept arguments.')
            return

        if not self.authenticated:
            if self.config.COMMANDS[cmd]['auth'] or (cmd == 'STAT' and arg):
                self.respond(b'530 Log in with USER and PASS first.')
                return
            else:
                # call the proper do_* method
                self._process_command(cmd, arg)
                return
        else:
            if (cmd == 'STAT') and not arg:
                self.do_STAT(path=None)
                return

            # for file-system related commands check whether real path
            # destination is valid
            if self.config.COMMANDS[cmd]['perm'] and (cmd != 'STOU'):
                if cmd in ('CWD', 'XCWD'):
                    if arg and self.working_dir != '/':
                        arg = os.path.join(self.working_dir, arg)
                    else:
                        arg = (arg or '/')
                elif cmd in ('CDUP', 'XCUP'):
                    arg = ''
                elif cmd == 'STAT':
                    if glob.has_magic(arg):
                        self.respond(b'550 Globbing not supported.')
                        return
                    arg = self.ftp_path(arg or self.working_dir)
                elif cmd == 'SITE CHMOD':
                    if ' ' not in arg:
                        self.respond(b'501 Syntax error: command needs two arguments.')
                        return
                    else:
                        mode, arg = arg.split(' ', 1)
                        arg = self.ftp_path(arg)
                        kwargs = dict(mode=mode)
                else:
                    if cmd == 'LIST':
                        if arg.lower() in ('-a', '-l', '-al', '-la'):
                            arg = self.working_dir
                        else:
                            arg = (arg or self.working_dir)
                    if glob.has_magic(arg):
                        self.respond(b'550 Globbing not supported.')
                        return
                    else:
                        arg = glob.escape(arg)
                        arg = (arg or self.working_dir)
                        arg = line.split(' ', 1)[1] if arg is None else arg

            # call the proper do_* method
            self._process_command(cmd, arg, **kwargs)

    def _process_command(self, cmd, *args, **kwargs):
        """Process command by calling the corresponding do_* class method (e.g. for received command "MKD pathname",
        do_MKD() method is called with "pathname" as the argument).
        """
        if self.invalid_login_attempt >= self.max_login_attempts:
            self.respond(b'421 Too many connections. Service temporarily unavailable.')
            self.disconnect_client = True
            self.session.add_event({'type': 'CONNECTION_TERMINATED'})
        else:
            try:
                method = getattr(self, 'do_' + cmd.replace(' ', '_'))
                self._last_command = cmd
                method(*args, **kwargs)
                if self._last_response:
                    code = int(self._last_response[:3])
                    resp = self._last_response[4:]
                    logger.debug('Last response {}:{} Client {}:{}'.format(code, resp, self.client_address,
                                                                           self.session.id))
            except (fs.errors.FSError, FilesystemError):
                raise

    # - main command processor
    def process_ftp_command(self):
        """
        Handle an incoming handle request - pick and item from the input_q, reads the contents of the message and
        dispatch contents to the appropriate do_* method.
        :param: (bytes) line - incoming request
        :return: (bytes) response - reply in respect to the request
        """
        try:
            if not self._command_channel_input_q.empty() and (self.metrics.timeout() < self.config.timeout):
                # decoding should be done using utf-8
                line = self._command_channel_input_q.get().decode()
                # Remove any CR+LF if present
                line = line[:-2] if line[-2:] == '\r\n' else line
                if line:
                    cmd = line.split(' ')[0].upper()
                    arg = line[len(cmd) + 1:]
                    try:
                        self._pre_process_cmd(line, cmd, arg)
                    except UnicodeEncodeError:
                        self.respond(b'501 can\'t decode path (server filesystem encoding is %a)' %
                                     sys.getfilesystemencoding())
                    except (fs.errors.PermissionDenied, FSOperationNotPermitted):
                        # TODO: log user as well.
                        logger.info('Client {} requested path: {} trying to access directory to which it has '
                                    'no access to.'.format(self.client_address, line))
                        self.respond(b'500 Permission denied')
                    except fs.errors.IllegalBackReference:
                        # Trying to access the directory which the current user has no access to
                        self.respond(b'550 %a points to a path which is outside the user\'s root directory.' %
                                     line)
                    except FTPPrivilegeException:
                        self.respond(b'550 Not enough privileges.')
                    except (fs.errors.FSError, FilesystemError) as fe:
                        logger.info('FTP client {} Unexpected error occurred : {}'.format(self.client_address, fe))
                        # TODO: what to respond here? For now just terminate the session
                        self.disconnect_client = True
                        self.session.add_event({'type': 'CONNECTION_TERMINATED'})
            elif not (self.metrics.timeout() < self.config.timeout) and (not self._data_channel):
                logger.info('FTP connection timeout, remote: {}. ({}). Disconnecting client'.format(self.client_address,
                                                                                                    self.session.id))
                self.session.add_event({'type': 'CONNECTION_TIMEOUT'})
                self.respond(b'421 Timeout.')
                self.disconnect_client = True
            else:
                gevent.sleep(0)

        except UnicodeDecodeError:
            # RFC-2640 doesn't mention what to do in this case. So we'll just return 501
            self.respond(b"501 can't decode command.")


# Copyright (C) 2017  Patrick Reichenberger (University of Passau) <patrick.reichenberger@t-online.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import gevent
import natsort
from conpot.protocols.IEC104.DeviceDataController import addr_in_hex, inro_response
from conpot.protocols.IEC104.i_frames_check import *
import conpot.core as conpot_core
from .frames import *

logger = logging.getLogger(__name__)


class IEC104(object):
    def __init__(self, device_data_controller, sock, address, session_id):
        self.sock = sock
        self.address = address
        self.session_id = session_id
        self.T_1 = conpot_core.get_databus().get_value('T_1')
        self.timeout_t1 = gevent.Timeout(self.T_1, gevent.Timeout)
        self.T_2 = conpot_core.get_databus().get_value('T_2')
        self.w = conpot_core.get_databus().get_value('w')
        self.device_data_controller = device_data_controller
        self.ssn = 0
        self.rsn = 0
        self.ack = 0
        self.allow_DT = False
        self.t2_caller = None
        self.telegram_count = 0
        self.sentmsgs = list()
        self.send_buffer = list()

    # === u_frame
    def handle_u_frame(self, frame):
        container = u_frame(frame)
        try:
            # check if valid u_frame (length, rest bits)
            if len(frame) == 6 and container.getfieldval("LenAPDU") == 4:
                if frame[3] == 0x00 and frame[4] == 0x00 and frame[5] == 0x00:
                    # check which type (Start, Stop, Test), only one active at same time
                    # STARTDT_act
                    if frame[2] == 0x07:
                        logger.info('%s ---> u_frame. STARTDT act. (%s)', self.address, self.session_id)
                        self.allow_DT = True
                        yield self.send_104frame(STARTDT_con)
                        # === If buffered data, send
                        if self.send_buffer:
                            for pkt in self.send_buffer:
                                yield self.send_104frame(pkt)
                    # STARTDT_con
                    elif frame[2] == 0x0B:
                        logger.info('%s ---> u_frame. STARTDT con. (%s)', self.address, self.session_id)
                        #  Station sends no STARTDT_act, so there is no STARTDT_con expected and no action performed
                        #  Can be extended, if used as Master
                    # STOPDT_act
                    elif frame[2] == 0x13:
                        logger.info('%s ---> u_frame. STOPDT act. (%s)', self.address, self.session_id)
                        self.allow_DT = False
                        # Send S_Frame
                        resp_frame = s_frame()
                        yield self.send_104frame(resp_frame)
                        yield self.send_104frame(STOPDT_con)
                    # STOPDT_con
                    elif frame[2] == 0x23:
                        logger.info('%s ---> u_frame. STOPDT con. (%s)', self.address, self.session_id)
                        self.timeout_t1.cancel()
                    # TESTFR_act
                    elif frame[2] == 0x43:
                        logger.info('%s ---> u_frame. TESTFR act. (%s)', self.address, self.session_id)
                        # In case of both sending a TESTFR_act.
                        if self.sentmsgs:
                            temp_list = []
                            for x in self.sentmsgs:
                                if x.name != 'u_frame' or x.getfieldval("Type") != 0x43:
                                    if isinstance(x, frame_object_with_timer):
                                        temp_list.append(x)
                                else:
                                    x.cancel_t1()
                            self.sentmsgs = temp_list
                        yield self.send_104frame(TESTFR_con)
                    # TESTFR_con
                    elif frame[2] == 0x83:
                        logger.info('%s ---> u_frame. TESTFR con. (%s)', self.address, self.session_id)
                        if self.sentmsgs:
                            temp_list = []
                            for x in self.sentmsgs:
                                if x.name != 'u_frame' or x.getfieldval("Type") != 0x43:
                                    if isinstance(x, frame_object_with_timer):
                                        temp_list.append(x)
                                else:
                                    x.cancel_t1()
                            self.sentmsgs = temp_list
                    else:
                        raise InvalidFieldValueException('Invalid u_frame packet, more than 1 bit set!  (%s)',
                                                         self.session_id)
                else:
                    raise InvalidFieldValueException('Control field octet 2,3 or 4 not 0x00! (%s)', self.session_id)
            else:
                raise InvalidFieldValueException('Wrong length for u_frame packet! (%s)', self.session_id)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s. (%s)", ex, self.session_id)

    # === s_frame
    def handle_s_frame(self, frame):
        container = s_frame(frame)
        try:
            # check if valid u_frame (length, rest bits)
            if len(frame) == 6 and container.getfieldval("LenAPDU") == 4:
                if frame[2] & 0x01 and frame[3] == 0x00:
                    recv_snr = container.getfieldval("RecvSeq")
                    logger.info("%s ---> s_frame receive nr: %s. (%s)",
                                self.address, str(recv_snr), self.session_id)
                    if recv_snr <= self.ssn:
                        self.ack = recv_snr
                    if self.sentmsgs:
                        temp_list = []
                        for x in self.sentmsgs:
                            if x.name != 'i_frame' or x.getfieldval("SendSeq") >= self.ack:
                                if isinstance(x, frame_object_with_timer):
                                    temp_list.append(x)
                            else:
                                x.cancel_t1()
                        self.sentmsgs = temp_list
                        self.show_send_list()

                else:
                    raise InvalidFieldValueException('Control field octet 1 in \'s_frame\' not 0x01 or 2 not 0x00! '
                                                     '(%s)', self.session_id)
            else:
                raise InvalidFieldValueException('Wrong length for s_frame packet! (%s)', self.session_id)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s. (%s)", ex, self.session_id)

    # === i_frame
    def handle_i_frame(self, frame):
        container = i_frame(frame)
        request_string = (" ".join(hex(n) for n in frame))
        logger.info("%s ---> i_frame: %s. (%s)", self.address, request_string, self.session_id)
        frame_length = len(frame)
        try:
            if container.getfieldval("LenAPDU") != frame_length - 2:
                raise InvalidFieldValueException('Wrong length for i_frame packet! (%s)', self.session_id)
            self.telegram_count += 1
            recv_snr = container.getfieldval("RecvSeq")

            # Figure 11
            if container.getfieldval("SendSeq") == self.rsn:
                self.recvseq_increment()
            else:
                logger.error("Sequence error, send s_frame for last correct packet. Then disconnect. (%s)",
                             self.session_id)
                # Better solution exists..
                if self.t2_caller:
                    gevent.kill(self.t2_caller)
                gevent.Greenlet.spawn_later(1, self.disconnect())
                return self.send_104frame(s_frame(RecvSeq=self.rsn))

            # All packets up to recv_snr-1 are acknowledged
            if self.sentmsgs:
                temp_list = []
                for x in self.sentmsgs:
                    if x.name != 'i_frame' or x.getfieldval("SendSeq") >= recv_snr:
                        if isinstance(x, frame_object_with_timer):
                            temp_list.append(x)
                    else:
                        x.cancel_t1()
                self.sentmsgs = temp_list

        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s. (%s)", ex, self.session_id)

        # Send S_Frame at w telegrams or (re)start timer T2
        resp_frame = s_frame()
        if not self.t2_caller:
            self.t2_caller = gevent.Greenlet.spawn_later(self.T_2, self.send_frame_imm, resp_frame)
        if self.telegram_count >= self.w:
            return self.send_104frame(resp_frame)

        type_id = container.getfieldval("TypeID")
        station_address = container.getfieldval("Addr")
        # 45: Single command
        if type_id == TypeIdentification["C_SC_NA_1"] and station_address == station_addr:
            return self.handle_single_command45(container)

        # 46: Double command
        elif type_id == TypeIdentification["C_DC_NA_1"] and station_address == station_addr:
            return self.handle_double_command46(container)

        # 49: Setpoint command, scaled value
        elif type_id == TypeIdentification["C_SE_NB_1"] and station_address == station_addr:
            return self.handle_setpointscaled_command49(container)

        # 50: Setpoint command, short floating point value
        elif type_id == TypeIdentification["C_SE_NC_1"] and station_address == station_addr:
            return self.handle_setpointfloatpoint_command50(container)

        # 100: (General-) Interrogation command
        elif type_id == TypeIdentification["C_IC_NA_1"] and station_address in (station_addr, 0xFFFF):
            return self.handle_inro_command100(container)

    def send_104frame(self, frame):
        # send s_frame
        if frame.name == "s_frame":
            frame.RecvSeq = self.rsn
            if self.t2_caller:
                gevent.kill(self.t2_caller)
            self.telegram_count = 0
            response_string = (" ".join(hex(n) for n in frame.build()))
            logger.info('%s <--- s_frame %s  (%s)', self.address, response_string, self.session_id)
            return frame.build()

        # send i_frame
        elif frame.name == "i_frame":
            if self.allow_DT:
                if self.t2_caller:
                    gevent.kill(self.t2_caller)
                frame.SendSeq = self.ssn
                frame.RecvSeq = self.rsn
                self.increment_sendseq()
                self.telegram_count = 0
                iframe = frame_object_with_timer(frame)
                self.sentmsgs.append(iframe)
                iframe.restart_t1()
                response_string = (" ".join(hex(n) for n in frame.build()))
                logger.info('%s <--- i_frame %s  (%s)', self.address, response_string, self.session_id)
                return frame.build()

            else:
                logger.info("StartDT missing, buffer data. (%s)", self.session_id)
                # Limitation for buffer, arbitrary number
                if len(self.send_buffer) < 50:
                    self.send_buffer.append(frame)

        # send u_frame
        elif frame.name == "u_frame":
            if frame.getfieldval("Type") == 0x07 or frame.getfieldval("Type") == 0x43:
                uframe = frame_object_with_timer(frame)
                self.sentmsgs.append(uframe)
                uframe.restart_t1()
            response_string = (" ".join(hex(n) for n in frame.build()))
            logger.info('%s <--- u_frame %s  (%s)', self.address, response_string, self.session_id)
            return frame.build()

    def send_frame_imm(self, frame):
        # send s_frame
        if frame.name == "s_frame":
            frame.RecvSeq = self.rsn
            if self.t2_caller:
                gevent.kill(self.t2_caller)
            self.telegram_count = 0
            response_string = (" ".join(hex(n) for n in frame.build()))
            logger.info('%s <--- s_frame %s  (%s)', self.address, response_string, self.session_id)
            return self.sock.send(frame.build())

    def handle_single_command45(self, container):
        try:
            check_asdu_45(container, "c")
            cause_of_transmission = int(container.getfieldval("COT"))
            info_obj_addr = container.getfieldval("IOA")
            field_val = container.getfieldval("SCS")
            if cause_of_transmission == 6:
                obj = self.device_data_controller.get_object_from_reg(info_obj_addr)  # get destination object
                if not (obj is None):  # if exists in xml-file
                    obj_cat = int(obj.category_id)  # get type (single command)
                    if obj_cat == 45:  # if object has type single command
                        # === Activation confirmation
                        act_con = i_frame() / asdu_head(COT=7) / asdu_infobj_45(IOA=info_obj_addr, SCS=field_val)
                        check_asdu_45(act_con, "m")
                        yield self.send_104frame(act_con)

                        # === Get related info object if exists
                        obj_rel_addr = obj.relation
                        if obj_rel_addr != "":  # if relation available
                            obj_rel_addr_hex = addr_in_hex(obj_rel_addr)  # get single point object address
                            # get the single point object
                            obj_rel = self.device_data_controller.get_object_from_reg(obj_rel_addr_hex)
                            obj.val = field_val  # set the value in the object to the command value
                            obj_rel.val = field_val  # set the value in the relation object to the command value
                            # test whether if it really updated the value
                            changed_val = obj_rel.val
                            single_point = i_frame() / asdu_head(COT=11) / asdu_infobj_1(IOA=obj_rel_addr_hex)
                            single_point.SIQ = SIQ(SPI=changed_val)
                            yield self.send_104frame(single_point)

                        # === Activation termination
                        act_term = i_frame() / asdu_head(COT=10) / asdu_infobj_45(IOA=info_obj_addr, SCS=field_val)
                        check_asdu_45(act_term, "m")
                        yield self.send_104frame(act_term)
                    else:  # if command type doesn't fit
                        # === neg. Activation confirmation
                        act_con = i_frame() / asdu_head(PN=1, COT=7) / asdu_infobj_45(IOA=info_obj_addr,
                                                                                      SCS=field_val)
                        check_asdu_45(act_con, "m")
                        yield self.send_104frame(act_con)
                else:  # object doesn't exist in xml file
                    # === unknown info obj address, object not found (or no reply?)
                    bad_addr = i_frame() / asdu_head(COT=47) / asdu_infobj_45(IOA=info_obj_addr, SCS=field_val)
                    check_asdu_45(bad_addr, "m")
                    yield self.send_104frame(bad_addr)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s  (%s)", ex, self.session_id)
        except AttributeError as ex:
            logger.warning("Allocation for field %s not possible. (%s)", ex, self.session_id)

    def handle_double_command46(self, container):
        try:
            check_asdu_46(container, "c")
            cause_of_transmission = int(container.getfieldval("COT"))
            info_obj_addr = container.getfieldval("IOA")
            field_val = container.getfieldval("DCS")
            if cause_of_transmission == 6:
                obj = self.device_data_controller.get_object_from_reg(info_obj_addr)  # get destination object
                if not (obj is None):  # if exists in xml-file
                    obj_cat = int(obj.category_id)  # get type (double command)
                    if obj_cat == 46:  # if object has type double command
                        # === Activation confirmation
                        act_con = i_frame() / asdu_head(COT=7) / asdu_infobj_46(IOA=info_obj_addr, DCS=field_val)
                        check_asdu_46(act_con, "m")
                        yield self.send_104frame(act_con)

                        # === Get related info object if exists
                        obj_rel_addr = obj.relation
                        if obj_rel_addr != "":  # if relation available
                            obj_rel_addr_hex = addr_in_hex(obj_rel_addr)  # get double point object address
                            # get the double point object
                            obj_rel = self.device_data_controller.get_object_from_reg(obj_rel_addr_hex)
                            obj.val = field_val  # set the value in the object to the command value
                            obj_rel.val = field_val  # set the value in the relation object to the command value
                            # test whether if it really updated the value
                            changed_val = obj_rel.val
                            double_point = i_frame() / asdu_head(COT=11) / asdu_infobj_3(IOA=obj_rel_addr_hex)
                            double_point.DIQ = DIQ(DPI=changed_val)
                            yield self.send_104frame(double_point)

                        # === Activation termination
                        act_term = i_frame() / asdu_head(COT=10) / asdu_infobj_46(IOA=info_obj_addr, DCS=field_val)
                        check_asdu_46(act_term, "m")
                        yield self.send_104frame(act_term)
                    else:  # if command type doesn't fit
                        # === neg. Activation confirmation
                        act_con = i_frame() / asdu_head(PN=1, COT=7) / asdu_infobj_46(IOA=info_obj_addr,
                                                                                      DCS=field_val)
                        check_asdu_46(act_con, "m")
                        yield self.send_104frame(act_con)
                else:  # object doesn't exist in xml file
                    # === unknown info obj address, object not found
                    bad_addr = i_frame() / asdu_head(COT=47) / asdu_infobj_46(IOA=info_obj_addr, DCS=field_val)
                    check_asdu_46(bad_addr, "m")
                    yield self.send_104frame(bad_addr)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s  (%s)", ex, self.session_id)
        except AttributeError as ex:
            logger.warning("Allocation for field %s not possible. (%s)", ex, self.session_id)

    def handle_setpointscaled_command49(self, container):
        try:
            check_asdu_49(container, "c")
            cause_of_transmission = int(container.getfieldval("COT"))
            info_obj_addr = container.getfieldval("IOA")
            field_val = container.getfieldval("SVA")
            if cause_of_transmission == 6:
                obj = self.device_data_controller.get_object_from_reg(info_obj_addr)  # get destination object
                if not (obj is None):  # if exists in xml-file
                    obj_cat = int(obj.category_id)  # get type (double command)
                    if obj_cat == 49:  # if object has type double command
                        # === Activation confirmation
                        act_con = i_frame() / asdu_head(COT=7) / asdu_infobj_49(IOA=info_obj_addr, SVA=field_val)
                        check_asdu_49(act_con, "m")
                        yield self.send_104frame(act_con)

                        # === Get related info object if exists
                        obj_rel_addr = obj.relation
                        if obj_rel_addr != "":  # if relation available
                            obj_rel_addr_hex = addr_in_hex(obj_rel_addr)  # get double point object address
                            # get the double point object
                            obj_rel = self.device_data_controller.get_object_from_reg(obj_rel_addr_hex)
                            obj.val = field_val  # set the value in the object to the command value
                            obj_rel.val = field_val  # set the value in the relation object to the command value
                            # test whether if it really updated the value
                            changed_val = obj_rel.val
                            setpoint_scaled = i_frame() / asdu_head(COT=3) / asdu_infobj_11(IOA=obj_rel_addr_hex,
                                                                                            SVA=changed_val)
                            setpoint_scaled.show2()
                            yield self.send_104frame(setpoint_scaled)

                        # === Activation termination
                        act_term = i_frame() / asdu_head(COT=10) / asdu_infobj_49(IOA=info_obj_addr, SVA=field_val)
                        check_asdu_49(act_term, "m")
                        yield self.send_104frame(act_term)
                    else:  # if command type doesn't fit
                        # === neg. Activation confirmation
                        act_con = i_frame() / asdu_head(PN=1, COT=7) / asdu_infobj_49(IOA=info_obj_addr, SVA=field_val)
                        check_asdu_49(act_con, "m")
                        yield self.send_104frame(act_con)
                else:  # object doesn't exist in xml file
                    # === unknown info obj address, object not found
                    bad_addr = i_frame() / asdu_head(COT=47) / asdu_infobj_49(IOA=info_obj_addr, SVA=field_val)
                    check_asdu_49(bad_addr, "m")
                    yield self.send_104frame(bad_addr)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s  (%s)", ex, self.session_id)
        except AttributeError as ex:
            logger.warning("Allocation for field %s not possible. (%s)", ex, self.session_id)

    def handle_setpointfloatpoint_command50(self, container):
        try:
            check_asdu_50(container, "c")
            cause_of_transmission = int(container.getfieldval("COT"))
            info_obj_addr = container.getfieldval("IOA")
            field_val = container.getfieldval("FPNumber")
            if cause_of_transmission == 6:
                obj = self.device_data_controller.get_object_from_reg(info_obj_addr)  # get destination object
                if not (obj is None):  # if exists in xml-file
                    obj_cat = int(obj.category_id)  # get type (double command)
                    if obj_cat == 50:  # if object has type double command
                        # === Activation confirmation
                        act_con = i_frame() / asdu_head(COT=7) / asdu_infobj_50(IOA=info_obj_addr, FPNumber=field_val)
                        check_asdu_50(act_con, "m")
                        yield self.send_104frame(act_con)

                        # === Get related info object if exists
                        obj_rel_addr = obj.relation
                        if obj_rel_addr != "":  # if relation available
                            obj_rel_addr_hex = addr_in_hex(obj_rel_addr)  # get double point object address
                            # get the double point object
                            obj_rel = self.device_data_controller.get_object_from_reg(obj_rel_addr_hex)
                            obj.val = field_val  # set the value in the object to the command value
                            obj_rel.val = field_val  # set the value in the relation object to the command value
                            # test whether if it really updated the value
                            changed_val = obj_rel.val
                            setpoint_scaled = i_frame() / asdu_head(COT=3) / asdu_infobj_13(IOA=obj_rel_addr_hex,
                                                                                            FPNumber=changed_val)
                            yield self.send_104frame(setpoint_scaled)

                        # === Activation termination
                        act_term = i_frame() / asdu_head(COT=10) / asdu_infobj_50(IOA=info_obj_addr, FPNumber=field_val)
                        check_asdu_50(act_term, "m")
                        yield self.send_104frame(act_term)
                    else:  # if command type doesn't fit
                        # === neg. Activation confirmation
                        act_con = i_frame() / asdu_head(PN=1, COT=7) / asdu_infobj_50(IOA=info_obj_addr,
                                                                                      FPNumber=field_val)
                        check_asdu_50(act_con, "m")
                        yield self.send_104frame(act_con)
                else:  # object doesn't exist in xml file
                    # === unknown info obj address, object not found
                    bad_addr = i_frame() / asdu_head(COT=47) / asdu_infobj_50(IOA=info_obj_addr, FPNumber=field_val)
                    check_asdu_50(bad_addr, "m")
                    yield self.send_104frame(bad_addr)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s  (%s)", ex, self.session_id)
        except AttributeError as ex:
            logger.warning("Allocation for field %s not possible. (%s)", ex, self.session_id)

    def handle_inro_command100(self, container):
        try:
            # check_asdu_100(container, "c")
            cause_of_transmission = container.getfieldval("COT")
            qualif_of_inro = container.getfieldval("QOI")
            if cause_of_transmission == 6:
                # === Activation confirmation for inro
                act_con_inro = i_frame() / asdu_head(COT=7) / asdu_infobj_100(QOI=qualif_of_inro)
                check_asdu_100(act_con_inro, "m")
                yield self.send_104frame(act_con_inro)
                # === Inro response
                if qualif_of_inro == 20:
                    reg = self.device_data_controller.get_registers()
                    sorted_reg = natsort.natsorted(list(reg.items()))

                    # get response list for certain types
                    resp1_list = inro_response(sorted_reg, 1)
                    resp3_list = inro_response(sorted_reg, 3)
                    resp5_list = inro_response(sorted_reg, 5)
                    resp7_list = inro_response(sorted_reg, 7)
                    resp9_list = inro_response(sorted_reg, 9)
                    resp11_list = inro_response(sorted_reg, 11)
                    resp13_list = inro_response(sorted_reg, 13)

                    # send each packet from each list
                    for resp1 in resp1_list:
                        yield self.send_104frame(resp1)
                    for resp3 in resp3_list:
                        yield self.send_104frame(resp3)
                    for resp5 in resp5_list:
                        yield self.send_104frame(resp5)
                    for resp7 in resp7_list:
                        yield self.send_104frame(resp7)
                    for resp9 in resp9_list:
                        yield self.send_104frame(resp9)
                    for resp11 in resp11_list:
                        yield self.send_104frame(resp11)
                    for resp13 in resp13_list:
                        yield self.send_104frame(resp13)

                # === Activation termination
                act_term = i_frame() / asdu_head(COT=10) / asdu_infobj_100(QOI=qualif_of_inro)
                check_asdu_100(act_con_inro, "m")
                yield self.send_104frame(act_term)
        except InvalidFieldValueException as ex:
            logger.warning("InvalidFieldValue: %s  (%s)", ex, self.session_id)
        except AttributeError as ex:
            logger.warning("Allocation for field %s not possible.  (%s)", ex, self.session_id)

    def restart_t1(self):
        self.timeout_t1.cancel()
        self.timeout_t1 = gevent.Timeout(self.T_1, gevent.Timeout)
        self.timeout_t1.start()

    def show_send_list(self):
        list_temp = list()
        for frm in self.sentmsgs:
            if frm.name == 'u_frame':
                u_type = str(hex(frm.getfieldval("Type")))
                list_temp.append("u(" + u_list[u_type] + ")")
            elif frm.name == 's_frame':
                s_type = frm.getfieldval("RecvSeq")
                list_temp.append("s(" + str(s_type) + ")")
            elif frm.name == 'i_frame':
                i_send_seq = frm.getfieldval("SendSeq")
                i_recv_seq = frm.getfieldval("RecvSeq")
                list_temp.append("i(" + str(i_send_seq) + "," + str(i_recv_seq) + ")")
        print(list_temp)

    def disconnect(self):
        self.timeout_t1.cancel()
        if self.t2_caller:
            gevent.kill(self.t2_caller)
        self.sock.close()
        self.ssn = 0
        self.rsn = 0
        self.ack = 0
        self.telegram_count = 0

    def increment_sendseq(self):
        if self.ssn < 65534:
            self.ssn = self.ssn + 2
        else:
            self.ssn = 0

    def recvseq_increment(self):
        if self.rsn < 65534:
            self.rsn = self.rsn + 2
        else:
            self.rsn = 0

    # returns object list for SQ = 0
    @staticmethod
    def get_infoobj_list(frame):
        info_obj_list = []
        number_of_objects = frame.getfieldval("NoO")
        for i in range(3, number_of_objects + 3):
            info_obj_list.append(frame.getlayer(i))
        return info_obj_list


class frame_object_with_timer:
    def __init__(self, frame):
        self.frame = frame
        self.name = frame.name
        self.T_1 = conpot_core.get_databus().get_value('T_1')
        self.__timeout_t1 = gevent.Timeout(self.T_1, gevent.Timeout)

    def restart_t1(self):
        self.__timeout_t1.cancel()
        self.__timeout_t1 = gevent.Timeout(self.T_1, gevent.Timeout)
        self.__timeout_t1.start()

    def cancel_t1(self):
        self.__timeout_t1.cancel()

    def getfieldval(self, fieldval):
        return self.frame.getfieldval(fieldval)

    def build(self):
        return self.frame.build()



# Copyright (C) 2014 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import json
import inspect
# this is needed because we use it in the xml.
import random

import gevent
import gevent.event
from lxml import etree


logger = logging.getLogger(__name__)


class Databus(object):
    def __init__(self):
        self._data = {}
        self._observer_map = {}
        self.initialized = gevent.event.Event()

    # the idea here is that we can store both values and functions in the key value store
    # functions could be used if a profile wants to simulate a sensor, or the function
    # could interface with a real sensor
    def get_value(self, key):
        logger.debug('DataBus: Get value from key: [%s]', key)
        assert key in self._data
        item = self._data[key]
        if getattr(item, "get_value", None):
            # this could potentially generate a context switch, but as long the called method
            # does not "callback" the databus we should be fine
            logger.debug('(K, V): (%s, %s)' % (key, item.get_value()))
            return item.get_value()
        elif hasattr(item, '__call__'):
            return item()
        else:
            # guaranteed to not generate context switch
            logger.debug('(K, V): (%s, %s)' % (key, item))
            return item

    def set_value(self, key, value):
        logger.debug('DataBus: Storing key: [%s] value: [%s]', key, value)
        self._data[key] = value
        # notify observers
        if key in self._observer_map:
            gevent.spawn(self.notify_observers, key)

    def notify_observers(self, key):
        for cb in self._observer_map[key]:
            cb(key)

    def observe_value(self, key, callback):
        assert hasattr(callback, '__call__')
        assert len(inspect.getargspec(callback)[0])  # depreciated in py3.5, un-depreciated in py3.6
        if key not in self._observer_map:
            self._observer_map[key] = []
        self._observer_map[key].append(callback)

    def initialize(self, config_file):
        self.reset()
        assert self.initialized.isSet() is False
        logger.debug('Initializing databus using %s.', config_file)
        dom = etree.parse(config_file)
        entries = dom.xpath('//core/databus/key_value_mappings/*')
        for entry in entries:
            key = entry.attrib['name']
            value = entry.xpath('./value/text()')[0].strip()
            value_type = str(entry.xpath('./value/@type')[0])
            assert key not in self._data
            logging.debug('Initializing %s with %s as a %s.', key, value, value_type)
            if value_type == 'value':
                self.set_value(key, eval(value))
            elif value_type == 'function':
                namespace, _classname = value.rsplit('.', 1)
                params = entry.xpath('./value/@param')
                module = __import__(namespace, fromlist=[_classname])
                _class = getattr(module, _classname)
                if len(params) > 0:
                    # eval param to list
                    params = eval(params[0])
                    self.set_value(key, _class(*(tuple(params))))
                else:
                    self.set_value(key, _class())
            else:
                raise Exception('Unknown value type: {0}'.format(value_type))
        self.initialized.set()

    def get_shapshot(self):
        # takes a snapshot of the internal honeypot state and returns it as json.
        snapsnot = {}
        for key in list(self._data.keys()):
            snapsnot[key] = self.get_value(key)
        return json.dumps(snapsnot)

    def reset(self):
        logger.debug('Resetting databus.')

        # if the class has a stop method call it.
        for value in list(self._data.values()):
            if getattr(value, "stop", None):
                value.stop()

        self._data.clear()
        self._observer_map.clear()
        self.initialized.clear()



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Utils related to ConpotVFS
"""
import fs
from typing import Optional, Union
from fs import errors
from fs.permissions import Permissions
import typing
from fs.subfs import SubFS
from fs.error_tools import unwrap_errors
import logging
_F = typing.TypeVar('_F', bound='FS', covariant=True)

logger = logging.getLogger(__name__)


class FilesystemError(fs.errors.FSError):
    """Custom class for filesystem-related exceptions."""


class FSOperationNotPermitted(fs.errors.FSError):
    """Custom class for filesystem-related exceptions."""


def copy_files(source, dest, buffer_size=1024*1024):
    """
    Copy a file from source to dest. source and dest must be file-like objects.
    """
    while True:
        copy_buffer = source.read(buffer_size)
        if not copy_buffer:
            break
        dest.write(copy_buffer)


class _custom_conpot_file(object):

    def __init__(self,
                 file_system,
                 parent_fs,
                 path,
                 mode,
                 buffering=-1,
                 encoding=None,
                 newline='',
                 line_buffering=False):
        self.file_system = file_system
        self._path = path
        self._file = parent_fs.open(path=self._path,
                                    mode=mode,
                                    buffering=buffering,
                                    encoding=encoding,
                                    newline=newline,
                                    line_buffering=line_buffering)
        self.mode = self._file.mode

    def __getattr__(self, item):
        return getattr(self._file, item)

    def __repr__(self):
        return '<conpot_fs cached file: {}>'.format(self._file.__repr__())

    __str__ = __repr__

    @property
    def get_file(self):
        return self._file

    def close(self):
        self._file.close()
        if ('w' in self.mode) or ('a' in self.mode) or (self.file_system.built_cache is False) or ('x' in self.mode):
            self.file_system._cache.update({self._path: self.file_system.getinfo(self._path, get_actual=True,
                                                                                 namespaces=['basic', 'access',
                                                                                             'details', 'stat'])})
            self.file_system.chown(self._path, self.file_system.default_uid, self.file_system.default_gid)
            self.file_system.chmod(self._path, self.file_system.default_perms)
        logger.debug('Updating modified/access time')
        self.file_system.setinfo(self._path, {})

    def __enter__(self):
        return self._file

    def __exit__(self, exc_type, exc_value, traceback):
        logger.debug('Exiting file at : {}'.format(self._path))
        self.close()


class SubAbstractFS(SubFS[_F], typing.Generic[_F]):
    """
    Creates a chroot jail sub file system. Each protocol can have an instance of this class. Use AbstractFS's
    create_jail method to access this. You won't be able to cd into an `up` directory.
    """
    def __init__(self, parent_fs, path):
        self.parent_fs = parent_fs
        self._default_uid, self._default_gid = parent_fs.default_uid, parent_fs.default_gid
        self._default_perms = parent_fs.default_perms
        self.utime = self.settimes
        super(SubAbstractFS, self).__init__(parent_fs, path)

    def getinfo(self, path: str, get_actual: bool = False, namespaces=None):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.getinfo(_path, get_actual=get_actual, namespaces=namespaces)

    # ------- Setters and getters for default users/grps/perms

    @property
    def default_perms(self):
        return self._default_perms

    @default_perms.setter
    def default_perms(self, perms):
        try:
            assert isinstance(perms, Permissions)
            self._default_perms = perms
        except AssertionError:
            raise FilesystemError('Permissions provided must be of valid type (fs.permissions.Permission)')

    @property
    def default_uid(self):
        return self._default_uid

    @default_uid.setter
    def default_uid(self, _uid):
        if _uid in self.parent_fs._users.keys():
            self._default_uid = _uid
        else:
            raise FilesystemError('User with id {} not registered with fs'.format(_uid))

    @property
    def default_gid(self):
        return self._default_gid

    @default_gid.setter
    def default_gid(self, _gid):
        if _gid in self.parent_fs._grps.keys():
            self._default_gid = _gid
        else:
            raise FilesystemError('Group with id {} not registered with fs'.format(_gid))

    # ---- Other utilites

    @property
    def default_user(self):
        return self.parent_fs._users[self.default_uid]['user']

    @property
    def default_group(self):
        return self.parent_fs._grps[self.default_gid]['group']

    def getcwd(self):
        return self._sub_dir

    @property
    def root(self):
        return self.parent_fs.root + self.getcwd()

    def getmtime(self, path):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.getmtime(_path)

    def format_list(self, basedir, listing):
        _fs, _path = self.delegate_path(basedir)
        with unwrap_errors(basedir):
            return _fs.format_list(_path, listing)

    def check_access(self, path=None, user=None, perms=None):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.check_access(_path, user, perms)

    def chown(self, fs_path: str, uid: int, gid: int, recursive: Optional[bool]=False):
        _fs, _path = self.delegate_path(fs_path)
        with unwrap_errors(fs_path):
            return _fs.chown(_path, uid, gid, recursive)

    def chmod(self, path: str, mode: oct, recursive: bool = False) -> None:
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.chmod(_path, mode, recursive)

    def access(self, path: str, name_or_id: Union[int, str]=None, required_perms: str=None):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.access(_path, name_or_id, required_perms)

    def stat(self, path):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.stat(_path)

    def readlink(self, path):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.readlink(_path)

    def get_permissions(self, path):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.get_permissions(_path)

    def removedir(self, path, rf=False):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.removedir(_path)

    def remove(self, path):
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.remove(_path)

    def move(self, src_path, dst_path, overwrite=True):
        _fs, _src_path = self.delegate_path(src_path)
        _, _dst_path = self.delegate_path(dst_path)
        with unwrap_errors({_src_path: src_path, _dst_path: dst_path}):
            return _fs.move(_src_path, _dst_path, overwrite=overwrite)

    def __getattr__(self, item):
        if hasattr(self.parent_fs, item) and item in {'_cache', 'create_group', 'register_user', 'take_snapshot',
                                                      'norm_path', 'users', 'groups', 'add_users_to_group',
                                                      'check_access'}:
            return getattr(self.parent_fs, item)
        else:
            raise NotImplementedError('Conpot\'s File System does not currently support method: {}'.format(item))


# Copyright (C) 2013  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gevent import monkey
import os
import unittest
from datetime import datetime
from collections import namedtuple
from gevent.queue import Queue
import conpot
from gevent.server import StreamServer
from conpot.protocols.modbus.slave import MBSlave
from modbus_tk.exceptions import ModbusInvalidResponseError, ModbusError
import modbus_tk.defines as cst
import modbus_tk.modbus_tcp as modbus_tcp
from gevent import socket
from conpot.protocols.modbus import modbus_server
import conpot.core as conpot_core
monkey.patch_all()


class TestModbusServer(unittest.TestCase):

    def setUp(self):
        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()

        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)

        # make paths platform-independent
        template = self.dir_name + '/templates/default/template.xml'
        modbus_template = self.dir_name + '/templates/default/modbus/modbus.xml'

        self.databus = conpot_core.get_databus()
        self.databus.initialize(template)
        args = namedtuple('FakeArgs', 'mibpaths raw_mib')
        self.modbus = modbus_server.ModbusServer(modbus_template, 'none', args, timeout=2)
        self.modbus_server = StreamServer(('127.0.0.1', 0), self.modbus.handle)
        self.modbus_server.start()

        # We have to use different slave IDs under different modes. In tcp mode,
        # only 255 and 0 make sense. However, modbus_tcp.TcpMaster explicitly
        # ignores slave ID 0. Therefore we can only use 255 in tcp mode.
        self.target_slave_id = 1 if self.modbus.mode == 'serial' else 255

    def tearDown(self):
        self.modbus_server.stop()

        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_read_coils(self):
        """
        Objective: Test if we can extract the expected bits from a slave using the modbus protocol.
        """

        self.databus.set_value('memoryModbusSlave%dBlockA' % self.target_slave_id, [1 for b in range(0, 128)])

        # create READ_COILS request
        master = modbus_tcp.TcpMaster(host='127.0.0.1', port=self.modbus_server.server_port)
        master.set_timeout(1.0)
        actual_bits = master.execute(slave=self.target_slave_id, function_code=cst.READ_COILS, starting_address=1,
                                     quantity_of_x=128)

        # the test template sets all bits to 1 in the range 1-128
        expected_bits = [1 for b in range(0, 128)]
        self.assertSequenceEqual(actual_bits, expected_bits)

    def test_write_read_coils(self):
        """
        Objective: Test if we can change values using the modbus protocol.
        """
        master = modbus_tcp.TcpMaster(host='127.0.0.1', port=self.modbus_server.server_port)
        master.set_timeout(1.0)
        set_bits = [1, 0, 0, 1, 0, 0, 1, 1]

        # write 8 bits
        master.execute(slave=self.target_slave_id, function_code=cst.WRITE_MULTIPLE_COILS, starting_address=1, output_value=set_bits)
        # read 8 bit
        actual_bit = master.execute(slave=self.target_slave_id, function_code=cst.READ_COILS, starting_address=1, quantity_of_x=8)

        self.assertSequenceEqual(set_bits, actual_bit)

    def test_read_nonexistent_slave(self):
        """
        Objective: Test if the correct exception is raised when trying to read from nonexistent slave.
        """
        master = modbus_tcp.TcpMaster(host='127.0.0.1', port=self.modbus_server.server_port)
        master.set_timeout(1.0)
        with self.assertRaises(ModbusError) as cm:
            master.execute(slave=5, function_code=cst.READ_COILS, starting_address=1, quantity_of_x=1)
        self.assertEqual(cm.exception.get_exception_code(), cst.SLAVE_DEVICE_FAILURE)

    def test_modbus_logging(self):
        """
        Objective: Test if modbus generates log messages as expected.
        Expected output is a dictionary with the following structure:
        {'timestamp': datetime.datetime(2013, 4, 23, 18, 47, 38, 532960),
         'remote': ('127.0.0.1', 60991),
         'data_type': 'modbus',
         'id': '01bd90d6-76f4-43cb-874f-5c8f254367f5',
         'data': {'function_code': 1,
                  'slave_id': 1,
                  'request': '0100010080',
                  'response': '0110ffffffffffffffffffffffffffffffff'}}
        """

        self.databus.set_value('memoryModbusSlave%dBlockA' % self.target_slave_id, [1 for b in range(0,128)])

        master = modbus_tcp.TcpMaster(host='127.0.0.1', port=self.modbus_server.server_port)
        master.set_timeout(1.0)

        # issue request to modbus server
        master.execute(slave=self.target_slave_id, function_code=cst.READ_COILS, starting_address=1, quantity_of_x=128)

        # extract the generated log entries
        log_queue = conpot_core.get_sessionManager().log_queue

        conn_log_item = log_queue.get(True, 2)
        conn_expected_payload = {'type': 'NEW_CONNECTION'}
        self.assertDictEqual(conn_expected_payload, conn_log_item['data'])

        modbus_log_item = log_queue.get(True, 2)
        self.assertIsInstance(modbus_log_item['timestamp'], datetime)
        self.assertTrue('data' in modbus_log_item)
        # we expect session_id to be 36 characters long (32 x char, 4 x dashes)
        self.assertTrue(len(str(modbus_log_item['id'])), modbus_log_item)
        self.assertEqual('127.0.0.1', modbus_log_item['remote'][0])
        self.assertEqual('modbus', modbus_log_item['data_type'])

        req = ('000100000006%s0100010080' % ('01' if self.target_slave_id == 1 else 'ff')).encode()
        # testing the actual modbus data
        modbus_expected_payload = {'function_code': 1, 'slave_id': self.target_slave_id,
                                   'request': req,
                                   'response': b'0110ffffffffffffffffffffffffffffffff'}

        self.assertDictEqual(modbus_expected_payload, modbus_log_item['data'])

    def test_report_slave_id(self):
        """
        Objective: Test conpot for function code 17.
        """
        # Function 17 is not currently supported by modbus_tk
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.modbus_server.server_port))
        s.sendall(b'\x00\x00\x00\x00\x00\x02\x01\x11')
        data = s.recv(1024)
        s.close()
        self.assertEqual(data, b'\x00\x00\x00\x00\x00\x06\x01\x11\x11\x01\x01\xff')

    def test_response_function_43_device_info(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.modbus_server.server_port))
        s.sendall(b'\x00\x01\x00\x00\x00\x05\x01\x2b\x0e\x01\x02')
        data = s.recv(1024)
        s.close()
        self.assertTrue(b'SIMATIC' in data and b'Siemens' in data)


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2014  Daniel creo Haslinger <creo-conpot@blackmesa.at>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
# from conpot.core.loggers.mysql_log import MySQLlogger


@unittest.skip("switch for a more generic db solution in py3")
class Test_MySQLlogger(unittest.TestCase):
    def test_mysqllogger(self):
        """
        Objective: Test if events can be stored to and retrieved from mysql properly.
        """

        # instanciate our mysql logging infrastructure
        host = '127.0.0.1'
        port = 3306
        username = 'travis'
        passphrase = ''
        db = 'conpot_unittest'
        logdevice = ''
        logsocket = 'tcp'
        sensorid = 'default'

        mysqllogger = MySQLlogger(host, port, db, username, passphrase, logdevice, logsocket, sensorid)

        # create a test event
        test_event = dict()
        test_event['id'] = 1337
        test_event['remote'] = "127.0.0.2"
        test_event['data_type'] = "unittest"
        test_event['data'] = {'request': 'foo', 'response': 'bar'}

        # lets do it, but do not retry in case of failure
        success = mysqllogger.log(test_event, 0)
        self.assertTrue(success, 'Could not log to mysql database')

        # now that we logged something, lets try to retrieve the event again..
        retrieved_event = mysqllogger.select_session_data(test_event['id'])
        self.assertEqual(len(retrieved_event), 1, 'Retrieved wrong number of events (or no event at all)')


# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent.monkey
gevent.monkey.patch_all()
import os
import unittest
from collections import namedtuple
import conpot
from conpot.protocols.s7comm.s7_server import S7Server
from conpot.tests.helpers import s7comm_client

import conpot.core as conpot_core


class TestS7Server(unittest.TestCase):

    def setUp(self):
        self.databus = conpot_core.get_databus()
        self.dir_name = os.path.dirname(conpot.__file__)
        self.databus.initialize(self.dir_name + '/templates/default/template.xml')
        args = namedtuple('FakeArgs', '')
        self.s7_instance = S7Server(self.dir_name + '/templates/default/s7comm/s7comm.xml', 'none', args)
        gevent.spawn(self.s7_instance.start, '127.0.0.1', 0)
        gevent.sleep(0.5)
        self.server_port = self.s7_instance.server.server_port

    def tearDown(self):
        self.s7_instance.stop()

    def test_s7(self):
        """
        Objective: Test if the S7 server returns the values expected.
        """
        src_tsaps = (0x100, 0x200)
        dst_tsaps = (0x102, 0x200, 0x201)
        s7_con = s7comm_client.s7('127.0.0.1', self.server_port)
        res = None
        for src_tsap in src_tsaps:
            for dst_tsap in dst_tsaps:
                try:
                    s7_con.src_tsap = src_tsap
                    s7_con.dst_tsap = dst_tsap
                    res = src_tsap, dst_tsap
                    break
                except s7comm_client.S7ProtocolError:
                    continue
            if res:
                break
        s7_con.src_ref = 10
        s7_con.s.settimeout(s7_con.timeout)
        s7_con.s.connect((s7_con.ip, s7_con.port))
        s7_con.Connect()
        identities = s7comm_client.GetIdentity('127.0.0.1', self.server_port, res[0], res[1])

        dic = {
            17: {1: "v.0.0"},
            28: {
                1: "Technodrome",
                2: "Siemens, SIMATIC, S7-200",
                3: "Mouser Factory",
                4: "Original Siemens Equipment",
                5: "88111222",
                7: "IM151-8 PN/DP CPU",
                10: "",
                11: ""
            }
        }

        for line in identities:
            sec, item, val = line.split(";")
            try:
                self.assertTrue(dic[int(sec)][int(item)] == val.strip())
            except AssertionError:
                print((sec, item, val))
                raise


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2013 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import ast
import textwrap

from mixbox import idgen
from mixbox.namespaces import Namespace

from stix.core import STIXHeader, STIXPackage
from stix.common import InformationSource
from stix.common.vocabs import VocabString
from stix.incident import Incident
from stix.incident.time import Time as StixTime
from stix.indicator import Indicator
from stix.ttp import TTP, VictimTargeting
from stix.extensions.identity.ciq_identity_3_0 import CIQIdentity3_0Instance, STIXCIQIdentity3_0, OrganisationInfo

from cybox.core import Observable
from cybox.objects.socket_address_object import SocketAddress
from cybox.objects.address_object import Address
from cybox.objects.port_object import Port
from cybox.objects.network_connection_object import NetworkConnection
from cybox.objects.artifact_object import Artifact, ZlibCompression, Base64Encoding
from cybox.common import ToolInformationList, ToolInformation
from cybox.common import Time as CyboxTime

from datetime import datetime

import conpot

CONPOT_NAMESPACE = 'mushmush-conpot'
CONPOT_NAMESPACE_URL = 'http://mushmush.org/conpot'


class StixTransformer(object):
    def __init__(self, config, dom):
        self.protocol_to_port_mapping = dict(
            modbus=502,
            snmp=161,
            http=80,
            s7comm=102,
        )
        port_path_list = ['//conpot_template/protocols/'+x+'/@port' for x in list(self.protocol_to_port_mapping.keys())]
        for port_path in port_path_list:
            try:
                protocol_port = ast.literal_eval(dom.xpath(port_path)[0])
                protocol_name = port_path.rsplit("/", 2)[1]
                self.protocol_to_port_mapping[protocol_name] = protocol_port
            except IndexError:
                continue
        conpot_namespace = Namespace(CONPOT_NAMESPACE_URL, CONPOT_NAMESPACE, '')
        idgen.set_id_namespace(conpot_namespace)

    def _add_header(self, stix_package, title, desc):
        stix_header = STIXHeader()
        stix_header.title = title
        stix_header.description = desc
        stix_header.information_source = InformationSource()
        stix_header.information_source.time = CyboxTime()
        stix_header.information_source.time.produced_time = datetime.now().isoformat()
        stix_package.stix_header = stix_header

    def transform(self, event):
        stix_package = STIXPackage()
        self._add_header(stix_package, "Unauthorized traffic to honeypot", "Describes one or more honeypot incidents")

        incident = Incident(id_="%s:%s-%s" % (CONPOT_NAMESPACE, 'incident', event['session_id']))
        initial_time = StixTime()
        initial_time.initial_compromise = event['timestamp'].isoformat()
        incident.time = initial_time
        incident.title = "Conpot Event"
        incident.short_description = "Traffic to Conpot ICS honeypot"
        incident.add_category(VocabString(value='Scans/Probes/Attempted Access'))

        tool_list = ToolInformationList()
        tool_list.append(ToolInformation.from_dict({
            'name': "Conpot",
            'vendor': "Conpot Team",
            'version': conpot.__version__,
            'description': textwrap.dedent('Conpot is a low interactive server side Industrial Control Systems '
                                           'honeypot designed to be easy to deploy, modify and extend.')
        }))
        incident.reporter = InformationSource(tools=tool_list)

        incident.add_discovery_method("Monitoring Service")
        incident.confidence = "High"

        # Victim Targeting by Sector
        ciq_identity = CIQIdentity3_0Instance()
        #identity_spec = STIXCIQIdentity3_0()
        #identity_spec.organisation_info = OrganisationInfo(industry_type="Electricity, Industrial Control Systems")
        #ciq_identity.specification = identity_spec
        ttp = TTP(title="Victim Targeting: Electricity Sector and Industrial Control System Sector")
        ttp.victim_targeting = VictimTargeting()
        ttp.victim_targeting.identity = ciq_identity

        incident.leveraged_ttps.append(ttp)

        indicator = Indicator(title="Conpot Event")
        indicator.description = "Conpot network event"
        indicator.confidence = "High"
        source_port = Port.from_dict({'port_value': event['remote'][1], 'layer4_protocol': 'tcp'})
        dest_port = Port.from_dict({'port_value': self.protocol_to_port_mapping[event['data_type']],
                                    'layer4_protocol': 'tcp'})
        source_ip = Address.from_dict({'address_value': event['remote'][0], 'category': Address.CAT_IPV4})
        dest_ip = Address.from_dict({'address_value': event['public_ip'], 'category': Address.CAT_IPV4})
        source_address = SocketAddress.from_dict({'ip_address': source_ip.to_dict(), 'port': source_port.to_dict()})
        dest_address = SocketAddress.from_dict({'ip_address': dest_ip.to_dict(), 'port': dest_port.to_dict()})
        network_connection = NetworkConnection.from_dict(
            {'source_socket_address': source_address.to_dict(),
             'destination_socket_address': dest_address.to_dict(),
             'layer3_protocol': "IPv4",
             'layer4_protocol': "TCP",
             'layer7_protocol': event['data_type'],
             'source_tcp_state': "ESTABLISHED",
             'destination_tcp_state': "ESTABLISHED",
             }
        )
        indicator.add_observable(Observable(network_connection))

        artifact = Artifact()
        artifact.data = json.dumps(event['data'])
        artifact.packaging.append(ZlibCompression())
        artifact.packaging.append(Base64Encoding())
        indicator.add_observable(Observable(artifact))

        incident.related_indicators.append(indicator)
        stix_package.add_incident(incident)

        stix_package_xml = stix_package.to_xml()
        return stix_package_xml
